# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""A plain CoAP resource directory according to
draft-ietf-core-resource-directory-25

Known Caveats:

    * It is very permissive. Not only is no security implemented.

    * This may and will make exotic choices about discoverable paths whereever
      it can (see StandaloneResourceDirectory documentation)

    * Split-horizon is not implemented correctly

    * Unless enforced by security (ie. not so far), endpoint and sector names
      (ep, d) are not checked for their lengths or other validity.

    * Simple registrations don't cache .well-known/core contents
"""

import string
import sys
import logging
import asyncio
import argparse
from urllib.parse import urljoin
import itertools

import aiocoap
from aiocoap.resource import Site, Resource, ObservableResource, PathCapable, WKCResource, link_format_to_message
from aiocoap.proxy.server import Proxy
from aiocoap.util.cli import AsyncCLIDaemon
import aiocoap.util.uri
from aiocoap import error
from aiocoap.cli.common import add_server_arguments, server_context_from_arguments
from aiocoap.numbers import codes, ContentFormat
import aiocoap.proxy.server

from aiocoap.util.linkformat import Link, LinkFormat, parse
from ..util.asyncio import py38args

from ..util.linkformat import link_header

IMMUTABLE_PARAMETERS = ('ep', 'd', 'proxy')

class NoActiveRegistration(error.ConstructionRenderableError):
    code = codes.PROXYING_NOT_SUPPORTED
    message = "no registration with that name"

def query_split(msg):
    """Split a message's query up into (key, [*value]) pairs from a
    ?key=value&key2=value2 style Uri-Query options.

    Keys without an `=` sign will have a None value, and all values are
    expressed as an (at least 1-element) list of repetitions.

    >>> m = aiocoap.Message(uri="coap://example.com/foo?k1=v1.1&k1=v1.2&obs")
    >>> query_split(m)
    {'k1': ['v1.1', 'v1.2'], 'obs': [None]}
    """
    result = {}
    for q in msg.opt.uri_query:
        if '=' not in q:
            k = q
            # matching the representation in link_header
            v = None
        else:
            k, v = q.split('=', 1)
        result.setdefault(k, []).append(v)
    return result

def pop_single_arg(query, name):
    """Out of query which is the output of query_split, pick the single value
    at the key name, raise a suitable BadRequest on error, or return None if
    nothing is there. The value is removed from the query dictionary."""

    if name not in query:
        return None
    if len(query[name]) > 1:
        raise error.BadRequest("Multiple values for %r" % name)
    return query.pop(name)[0]

class CommonRD:
    # "Key" here always means an (ep, d) tuple.

    entity_prefix = ("reg",)

    def __init__(self, proxy_domain=None, log=None):
        super().__init__()

        self.log = log or logging.getLogger('resource-directory')

        self._by_key = {} # key -> Registration
        self._by_path = {} # path -> Registration

        self._updated_state_cb = []

        self.proxy_domain = proxy_domain
        self.proxy_active = {} # uri_host -> Remote

    class Registration:
        # FIXME: split this into soft and hard grace period (where the former
        # may be 0). the node stays discoverable for the soft grace period, but
        # the registration stays alive for a (possibly much longer, at least
        # +lt) hard grace period, in which any action on the reg resource
        # reactivates it -- preventing premature reuse of the resource URI
        grace_period = 15

        @property
        def href(self):
            return '/' + '/'.join(self.path)

        def __init__(self, static_registration_parameters, path, network_remote, delete_cb, update_cb, registration_parameters, proxy_host, setproxyremote_cb):
            # note that this can not modify d and ep any more, since they are
            # already part of the key and possibly the path
            self.path = path
            self.links = LinkFormat([])

            self._delete_cb = delete_cb
            self._update_cb = update_cb

            self.registration_parameters = static_registration_parameters
            self.lt = 90000
            self.base_is_explicit = False

            self.proxy_host = proxy_host
            self._setproxyremote_cb = setproxyremote_cb

            self.update_params(network_remote, registration_parameters, is_initial=True)

        def update_params(self, network_remote, registration_parameters, is_initial=False):
            """Set the registration_parameters from the parsed query arguments,
            update any effects of them, and trigger any observation updates if
            required (the typical ones don't because their
            registration_parameters are {} and all it does is restart the
            lifetime counter)"""

            if any(k in ('ep', 'd') for k in registration_parameters.keys()):
                # The ep and d of initial registrations are already popped out
                raise error.BadRequest("Parameters 'd' and 'ep' can not be updated")

            # Not in use class "R" or otherwise conflict with common parameters
            if any(k in ('page', 'count', 'rt', 'href', 'anchor') for k in
                    registration_parameters.keys()):
                raise error.BadRequest("Unsuitable parameter for registration")

            if (is_initial or not self.base_is_explicit) and 'base' not in \
                    registration_parameters:
                # check early for validity to avoid side effects of requests
                # answered with 4.xx
                if self.proxy_host is None:
                    try:
                        network_base = network_remote.uri
                    except error.AnonymousHost:
                        raise error.BadRequest("explicit base required")
                else:
                    # FIXME: Advertise alternative transports (write alternative-transports)
                    network_base = 'coap://' + self.proxy_host

            if is_initial:
                # technically might be a re-registration, but we can't catch that at this point
                actual_change = True
            else:
                actual_change = False

            # Don't act while still checking
            set_lt = None
            set_base = None

            if 'lt' in registration_parameters:
                try:
                    set_lt = int(pop_single_arg(registration_parameters, 'lt'))
                except ValueError:
                    raise error.BadRequest("lt must be numeric")

            if 'base' in registration_parameters:
                set_base = pop_single_arg(registration_parameters, 'base')

            if set_lt is not None and self.lt != set_lt:
                actual_change = True
                self.lt = set_lt
            if set_base is not None and (is_initial or self.base != set_base):
                actual_change = True
                self.base = set_base
                self.base_is_explicit = True

            if not self.base_is_explicit and (is_initial or self.base != network_base):
                self.base = network_base
                actual_change = True

            if any(v != self.registration_parameters.get(k) for (k, v) in registration_parameters.items()):
                self.registration_parameters.update(registration_parameters)
                actual_change = True

            if is_initial:
                self._set_timeout()
            else:
                self.refresh_timeout()

            if actual_change:
                self._update_cb()

            if self.proxy_host:
                self._setproxyremote_cb(network_remote)

        def delete(self):
            self.timeout.cancel()
            self._update_cb()
            self._delete_cb()

        def _set_timeout(self):
            delay = self.lt + self.grace_period
            # workaround for python issue20493

            async def longwait(delay, callback):
                await asyncio.sleep(delay)
                callback()
            self.timeout = asyncio.create_task(
                    longwait(delay, self.delete),
                    **py38args(name="RD Timeout for %r" % self)
                    )

        def refresh_timeout(self):
            self.timeout.cancel()
            self._set_timeout()

        def get_host_link(self):
            attr_pairs = []
            for (k, values) in self.registration_parameters.items():
                for v in values:
                    attr_pairs.append([k, v])
            return Link(href=self.href, attr_pairs=attr_pairs, base=self.base, rt="core.rd-ep")

        def get_based_links(self):
            """Produce a LinkFormat object that represents all statements in
            the registration, resolved to the registration's base (and thus
            suitable for comparing anchors)."""
            result = []
            for l in self.links.links:
                href = urljoin(self.base, l.href)
                if 'anchor' in l:
                    absanchor = urljoin(self.base, l.anchor)
                    data = [(k, v) for (k, v) in l.attr_pairs if k != 'anchor'] + [['anchor', absanchor]]
                else:
                    data = l.attr_pairs + [['anchor', urljoin(href, '/')]]
                result.append(Link(href, data))
            return LinkFormat(result)

    async def shutdown(self):
        pass

    def register_change_callback(self, callback):
        """Ask RD to invoke the callback whenever any of the RD state
        changed"""
        # This has no unregister equivalent as it's only called by the lookup
        # resources that are expected to be live for the remainder of the
        # program, like the Registry is.
        self._updated_state_cb.append(callback)

    def _updated_state(self):
        for cb in self._updated_state_cb:
            cb()

    def _new_pathtail(self):
        for i in itertools.count(1):
            # In the spirit of making legal but unconvential choices (see
            # StandaloneResourceDirectory documentation): Whoever strips or
            # ignores trailing slashes shall have a hard time keeping
            # registrations alive.
            path = (str(i), '')
            if path not in self._by_path:
                return path

    def initialize_endpoint(self, network_remote, registration_parameters):
        # copying around for later use in static, but not checking again
        # because reading them from the original will already have screamed by
        # the time this is used
        static_registration_parameters = {k: v for (k, v) in registration_parameters.items() if k in IMMUTABLE_PARAMETERS}

        ep = pop_single_arg(registration_parameters, 'ep')
        if ep is None:
            raise error.BadRequest("ep argument missing")
        d = pop_single_arg(registration_parameters, 'd')

        proxy = pop_single_arg(registration_parameters, 'proxy')

        if proxy is not None and proxy != 'on':
            raise error.BadRequest("Unsupported proxy value")

        key = (ep, d)

        if static_registration_parameters.pop('proxy', None):
            # FIXME: 'ondemand' is done unconditionally

            if not self.proxy_domain:
                raise error.BadRequest("Proxying not enabled")

            def is_usable(s):
                # Host names per RFC1123 (which is stricter than what RFC3986 would allow).
                #
                # Only supporting lowercase names as to avoid ambiguities due
                # to hostname capitalizatio normalization (otherwise it'd need
                # to be first-registered-first-served)
                return s and all(x in string.ascii_lowercase + string.digits + '-' for x in s)
            if not is_usable(ep) or (d is not None and not is_usable(d)):
                raise error.BadRequest("Proxying only supported for limited ep and d set (lowercase, digits, dash)")

            proxy_host = ep
            if d is not None:
                proxy_host += '.' + d
            proxy_host = proxy_host + '.' + self.proxy_domain
        else:
            proxy_host = None

        # No more errors should fly out from below here, as side effects start now

        try:
            oldreg = self._by_key[key]
        except KeyError:
            path = self._new_pathtail()
        else:
            path = oldreg.path[len(self.entity_prefix):]
            oldreg.delete()

        # this was the brutal way towards idempotency (delete and re-create).
        # if any actions based on that are implemented here, they have yet to
        # decide wheter they'll treat idempotent recreations like deletions or
        # just ignore them unless something otherwise unchangeable (ep, d)
        # changes.

        def delete():
            del self._by_path[path]
            del self._by_key[key]
            self.proxy_active.pop(proxy_host, None)

        def setproxyremote(remote):
            self.proxy_active[proxy_host] = remote

        reg = self.Registration(static_registration_parameters, self.entity_prefix + path, network_remote, delete,
                self._updated_state, registration_parameters, proxy_host, setproxyremote)

        self._by_key[key] = reg
        self._by_path[path] = reg

        return reg

    def get_endpoints(self):
        return self._by_key.values()

def link_format_from_message(message):
    """Convert a response message into a LinkFormat object

    This expects an explicit media type set on the response (or was explicitly requested)
    """
    certain_format = message.opt.content_format
    if certain_format is None and hasattr(message, 'request'):
        certain_format = message.request.opt.accept
    try:
        if certain_format == ContentFormat.LINKFORMAT:
            return parse(message.payload.decode('utf8'))
        else:
            raise error.UnsupportedMediaType()
    except (UnicodeDecodeError, link_header.ParseException):
        raise error.BadRequest()

class ThingWithCommonRD:
    def __init__(self, common_rd):
        super().__init__()
        self.common_rd = common_rd

        if isinstance(self, ObservableResource):
            self.common_rd.register_change_callback(self.updated_state)

class DirectoryResource(ThingWithCommonRD, Resource):
    ct = link_format_to_message.supported_ct
    rt = "core.rd"

    #: Issue a custom warning when registrations come in via this interface
    registration_warning = None

    async def render_post(self, request):
        links = link_format_from_message(request)

        registration_parameters = query_split(request)

        if self.registration_warning:
            # Conveniently placed so it could be changed to something setting
            # additional registration_parameters instead
            self.common_rd.log.warning("Warning from registration: %s", self.registration_warning)

        regresource = self.common_rd.initialize_endpoint(request.remote, registration_parameters)
        regresource.links = links

        return aiocoap.Message(code=aiocoap.CREATED, location_path=regresource.path)

class RegistrationResource(Resource):
    """The resource object wrapping a registration is just a very thin and
    ephemeral object; all those methods could just as well be added to
    Registration with `s/self.reg/self/g`, making RegistrationResource(reg) =
    reg (or handleded in a single RegistrationDispatchSite), but this is kept
    here for better separation of model and interface."""

    def __init__(self, registration):
        super().__init__()
        self.reg = registration

    async def render_get(self, request):
        return link_format_to_message(request, self.reg.links)

    def _update_params(self, msg):
        query = query_split(msg)
        self.reg.update_params(msg.remote, query)

    async def render_post(self, request):
        self._update_params(request)

        if request.opt.content_format is not None or request.payload:
            raise error.BadRequest("Registration update with body not specified")

        return aiocoap.Message(code=aiocoap.CHANGED)

    async def render_put(self, request):
        # this is not mentioned in the current spec, but seems to make sense
        links = link_format_from_message(request)

        self._update_params(request)
        self.reg.links = links

        return aiocoap.Message(code=aiocoap.CHANGED)

    async def render_delete(self, request):
        self.reg.delete()

        return aiocoap.Message(code=aiocoap.DELETED)

class RegistrationDispatchSite(ThingWithCommonRD, Resource, PathCapable):
    async def render(self, request):
        try:
            entity = self.common_rd._by_path[request.opt.uri_path]
        except KeyError:
            raise error.NotFound

        entity = RegistrationResource(entity)

        return await entity.render(request.copy(uri_path=()))

def _paginate(candidates, query):
    page = pop_single_arg(query, 'page')
    count = pop_single_arg(query, 'count')

    try:
        candidates = list(candidates)
        if page is not None:
            candidates = candidates[int(page) * int(count):]
        if count is not None:
            candidates = candidates[:int(count)]
    except (KeyError, ValueError):
        raise error.BadRequest("page requires count, and both must be ints")

    return candidates

def _link_matches(link, key, condition):
    return any(k == key and condition(v) for (k, v) in link.attr_pairs)

class EndpointLookupInterface(ThingWithCommonRD, ObservableResource):
    ct = link_format_to_message.supported_ct
    rt = "core.rd-lookup-ep"

    async def render_get(self, request):
        query = query_split(request)

        candidates = self.common_rd.get_endpoints()

        for search_key, search_values in query.items():
            if search_key in ('page', 'count'):
                continue # filtered last

            for search_value in search_values:
                if search_value is not None and search_value.endswith('*'):
                    def matches(x, start=search_value[:-1]):
                        return x.startswith(start)
                else:
                    def matches(x, search_value=search_value):
                        return x == search_value

                if search_key in ('if', 'rt'):
                    def matches(x, original_matches=matches):
                        return any(original_matches(v) for v in x.split())

                if search_key == 'href':
                    candidates = (c for c in candidates if
                            matches(c.href) or
                            any(matches(r.href) for r in c.get_based_links().links)
                            )
                    continue

                candidates = (c for c in candidates if
                        (search_key in c.registration_parameters and any(matches(x) for x in c.registration_parameters[search_key])) or
                        any(_link_matches(r, search_key, matches) for r in c.get_based_links().links)
                        )

        candidates = _paginate(candidates, query)

        result = [c.get_host_link() for c in candidates]

        return link_format_to_message(request, LinkFormat(result))

class ResourceLookupInterface(ThingWithCommonRD, ObservableResource):
    ct = link_format_to_message.supported_ct
    rt = "core.rd-lookup-res"

    async def render_get(self, request):
        query = query_split(request)

        eps = self.common_rd.get_endpoints()
        candidates = ((e, c) for e in eps for c in e.get_based_links().links)

        for search_key, search_values in query.items():
            if search_key in ('page', 'count'):
                continue # filtered last

            for search_value in search_values:
                if search_value is not None and search_value.endswith('*'):
                    def matches(x, start=search_value[:-1]):
                        return x.startswith(start)
                else:
                    def matches(x, search_value=search_value):
                        return x == search_value

                if search_key in ('if', 'rt'):
                    def matches(x, original_matches=matches):
                        return any(original_matches(v) for v in x.split())

                if search_key == 'href':
                    candidates = ((e, c) for (e, c) in candidates if
                            matches(c.href) or
                            matches(e.href) # FIXME: They SHOULD give this as relative as we do, but don't have to
                            )
                    continue

                candidates = ((e, c) for (e, c) in candidates if
                        _link_matches(c, search_key, matches) or
                        (search_key in e.registration_parameters and any(matches(x) for x in e.registration_parameters[search_key]))
                        )

        # strip endpoint
        candidates = (c for (e, c) in candidates)

        candidates = _paginate(candidates, query)

        # strip needless anchors
        candidates = [
                Link(l.href, [(k, v) for (k, v) in l.attr_pairs if k != 'anchor'])
                if dict(l.attr_pairs)['anchor'] == urljoin(l.href, '/')
                else l
                for l in candidates]

        return link_format_to_message(request, LinkFormat(candidates))

class SimpleRegistration(ThingWithCommonRD, Resource):
    #: Issue a custom warning when registrations come in via this interface
    registration_warning = None

    def __init__(self, common_rd, context):
        super().__init__(common_rd)
        self.context = context

    async def render_post(self, request):
        query = query_split(request)

        if 'base' in query:
            raise error.BadRequest("base is not allowed in simple registrations")

        await self.process_request(
                network_remote=request.remote,
                registration_parameters=query,
            )

        return aiocoap.Message(code=aiocoap.CHANGED)

    async def process_request(self, network_remote, registration_parameters):
        if 'proxy' not in registration_parameters:
            try:
                network_base = network_remote.uri
            except error.AnonymousHost:
                raise error.BadRequest("explicit base required")

            fetch_address = (network_base + '/.well-known/core')
            get = aiocoap.Message(uri=fetch_address)
        else:
            # ignoring that there might be a based present, that will err later
            get = aiocoap.Message(uri_path=['.well-known', 'core'])
            get.remote = network_remote

        get.code = aiocoap.GET
        get.opt.accept = ContentFormat.LINKFORMAT

        # not trying to catch anything here -- the errors are most likely well renderable into the final response
        response = await self.context.request(get).response_raising
        links = link_format_from_message(response)

        if self.registration_warning:
            # Conveniently placed so it could be changed to something setting
            # additional registration_parameters instead
            self.common_rd.log.warning("Warning from registration: %s", self.registration_warning)
        registration = self.common_rd.initialize_endpoint(network_remote, registration_parameters)
        registration.links = links

class SimpleRegistrationWKC(WKCResource, SimpleRegistration):
    def __init__(self, listgenerator, common_rd, context):
        super().__init__(listgenerator=listgenerator, common_rd=common_rd, context=context)
        self.registration_warning = "via .well-known/core"

class StandaloneResourceDirectory(Proxy, Site):
    """A site that contains all function sets of the CoAP Resource Directoru

    To prevent or show ossification of example paths in the specification, all
    function set paths are configurable and default to values that are
    different from the specification (but still recognizable)."""

    rd_path = ("resourcedirectory", "")
    ep_lookup_path = ("endpoint-lookup", "")
    res_lookup_path = ("resource-lookup", "")

    def __init__(self, context, lwm2m_compat=None, **kwargs):
        if lwm2m_compat is True:
            self.rd_path = ("rd",)

        # Double inheritance: works as everything up of Proxy has the same interface
        super().__init__(outgoing_context=context)

        common_rd = CommonRD(**kwargs)

        self.add_resource([".well-known", "core"], SimpleRegistrationWKC(self.get_resources_as_linkheader, common_rd=common_rd, context=context))
        self.add_resource([".well-known", "rd"], SimpleRegistration(common_rd=common_rd, context=context))

        self.add_resource(self.rd_path, DirectoryResource(common_rd=common_rd))
        if list(self.rd_path) != ["rd"] and lwm2m_compat is None:
            second_dir_resource = DirectoryResource(common_rd=common_rd)
            second_dir_resource.registration_warning = "via unannounced /rd"
            # Hide from listing
            second_dir_resource.get_link_description = lambda *args: None
            self.add_resource(["rd"], second_dir_resource)
        self.add_resource(self.ep_lookup_path, EndpointLookupInterface(common_rd=common_rd))
        self.add_resource(self.res_lookup_path, ResourceLookupInterface(common_rd=common_rd))

        self.add_resource(common_rd.entity_prefix, RegistrationDispatchSite(common_rd=common_rd))

        self.common_rd = common_rd

    def apply_redirection(self, request):
        # Fully overriding so we don't need to set an add_redirector

        try:
            actual_remote = self.common_rd.proxy_active[request.opt.uri_host]
        except KeyError:
            self.common_rd.log.info("Request to proxied host %r rejected: No such registration", request.opt.uri_host)
            raise NoActiveRegistration
        self.common_rd.log.debug("Forwarding request to %r to remote %s", request.opt.uri_host, actual_remote)
        request.remote = actual_remote
        request.opt.uri_host = None
        return request

    async def shutdown(self):
        await self.common_rd.shutdown()

    async def render(self, request):
        # Full override switching which of the parents' behavior to choose

        if self.common_rd.proxy_domain is not None and \
                request.opt.uri_host is not None and \
                request.opt.uri_host.endswith("." + self.common_rd.proxy_domain): # in self.common_rd.proxy_active:
            return await Proxy.render(self, request)
        else:
            return await Site.render(self, request)

    # See render; necessary on all functions thanks to https://github.com/chrysn/aiocoap/issues/251

    async def needs_blockwise_assembly(self, request):
        if request.opt.uri_host in self.common_rd.proxy_active:
            return await Proxy.needs_blockwise_assembly(self, request)
        else:
            return await Site.needs_blockwise_assembly(self, request)

    async def add_observation(self, request, serverobservation):
        if request.opt.uri_host in self.common_rd.proxy_active:
            return await Proxy.add_observation(self, request, serverobservation)
        else:
            return await Site.add_observation(self, request, serverobservation)

def build_parser():
    p = argparse.ArgumentParser(description=__doc__)

    add_server_arguments(p)

    return p

class Main(AsyncCLIDaemon):
    async def start(self, args=None):
        parser = build_parser()
        parser.add_argument("--proxy-domain", help="Enable the RD proxy extension. Example: `proxy.example.net` will produce base URIs like `coap://node1.proxy.example.net/`. The names must all resolve to an address the RD is bound to.", type=str)
        parser.add_argument("--lwm2m-compat", help="Compatibility mode for LwM2M clients that can not perform some discovery steps (moving the registration resource to `/rd`)", action='store_true', default=None)
        parser.add_argument("--no-lwm2m-compat", help="Disable all compativility with LwM2M clients that can not perform some discovery steps (not even accepting registrations at `/rd` with warnings)", action='store_false', dest='lwm2m_compat')
        parser.add_argument("--verbose", help="Increase debug log output (repeat for increased verbosity)", action='count', default=0)
        options = parser.parse_args(args if args is not None else sys.argv[1:])

        # Putting in an empty site to construct the site with a context
        self.context = await server_context_from_arguments(None, options)

        self.log = logging.getLogger('resource-directory')
        if options.verbose >= 2:
            self.log.setLevel(logging.DEBUG)
        elif options.verbose == 1:
            self.log.setLevel(logging.INFO)

        self.site = StandaloneResourceDirectory(
                context=self.context,
                proxy_domain=options.proxy_domain,
                lwm2m_compat=options.lwm2m_compat,
                log=self.log,
                )
        self.context.serversite = self.site

    async def shutdown(self):
        await self.site.shutdown()
        await self.context.shutdown()

sync_main = Main.sync_main

if __name__ == "__main__":
    sync_main()



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import functools
import random

from . import error
from . import interfaces
# To be used sparingly here: This deals with request / responses on the token
# layer. But the layer below won't even know that messages are responses, so it
# can't make the informed decisions we make here.
from .numbers.types import NON
from .pipe import Pipe

class TokenManager(interfaces.RequestInterface, interfaces.TokenManager):

    def __init__(self, context):
        self.context = context

        self._token = random.randint(0, 65535)
        self.outgoing_requests = {}
        """Unfinished outgoing requests (identified by token and remote)"""
        self.incoming_requests = {}
        """Unfinished incoming requests.

        ``(token, remote): (Pipe, stopper)`` where stopper is a
        function unregistes the Pipe event handler and thus
        indicates to the server the discontinued interest"""

        self.log = self.context.log
        self.loop = self.context.loop

        #self.token_interface = … -- needs to be set post-construction, because the token_interface in its constructor already needs to get its manager

    def __repr__(self):
        return '<%s for %s>' % (type(self).__name__, getattr(self, 'token_interface', '(unbound)'))

    @property
    def client_credentials(self):
        return self.context.client_credentials

    async def shutdown(self):
        while self.incoming_requests:
            key = next(iter(self.incoming_requests.keys()))
            (_, stop) = self.incoming_requests.pop(key)
            # This cancels them, not sending anything.
            #
            # FIXME should we? (RST? 5.00 Server Shutdown? An RST would only
            # work if we pushed this further down the shutdown chain; a 5.00 we
            # could raise in the task.)
            stop()
        self.incoming_requests = None

        while self.outgoing_requests:
            key = next(iter(self.outgoing_requests.keys()))
            request = self.outgoing_requests.pop(key)
            request.add_exception(error.LibraryShutdown())
        self.outgoing_requests = None

        await self.token_interface.shutdown()

    def next_token(self):
        """Reserve and return a new Token for request."""
        #TODO: add proper Token handling
        self._token = (self._token + 1) % (2 ** 64)
        return self._token.to_bytes(8, 'big').lstrip(b'\0')

    #
    # implement the tokenmanager interface
    #

    def dispatch_error(self, exception, remote):
        if self.outgoing_requests is None:
            # Not entirely sure where it is so far; better just raise a warning
            # than an exception later, nothing terminally bad should come of
            # this error.
            self.log.warning("Internal shutdown sequence msismatch: error dispatched through tokenmanager after shutdown")
            return

        # NetworkError is what we promise users to raise from request etc; if
        # it's already a NetworkError and possibly more descriptive (eg. a
        # TimeoutError), we'll just let it through (and thus allow
        # differentiated handling eg. in application-level retries).
        if not isinstance(exception, error.NetworkError):
            cause = exception
            exception = error.NetworkError(str(exception))
            exception.__cause__ = cause

        # The stopping calls would pop items from the pending requests --
        # iterating once, extracting the stoppers and then calling them en
        # batch
        stoppers = []
        for key, request in self.outgoing_requests.items():
            (token, request_remote) = key
            if request_remote == remote:
                stoppers.append(lambda request=request, exception=exception: request.add_exception(exception))

        for ((_, _r), (_, stopper)) in self.incoming_requests.items():
            if remote == _r:
                stoppers.append(stopper)
        for stopper in stoppers:
            stopper()

    def process_request(self, request):
        key = (request.token, request.remote)

        if key in self.incoming_requests:
            # This is either a "I consider that token invalid, probably forgot
            # about it, but here's a new request" or renewed interest in an
            # observation, which gets modelled as a new request at thislevel
            self.log.debug("Incoming request overrides existing request")
            # Popping: FIXME Decide if one of them is sufficient (see `del self.incoming_requests[key]` below)
            (pipe, stop) = self.incoming_requests.pop(key)
            stop()

        pipe = Pipe(request, self.log)

        # FIXME: what can we pass down to the token_interface?  certainly not
        # the request, but maybe the request with a response filter applied?
        def on_event(ev):
            if ev.message is not None:
                m = ev.message
                # FIXME: should this code warn if token or remote are set?
                m.token = request.token
                m.remote = request.remote.as_response_address()

                if m.mtype is None and request.mtype is NON:
                    # Default to sending NON to NON requests; rely on the
                    # default (CON if stand-alone else ACK) otherwise.
                    m.mtype = NON
                self.token_interface.send_message(
                        m,
                        # No more interest from *that* remote; as it's the only
                        # thing keeping the PR alive, it'll go its course of
                        # vanishing for lack of interest (as it would if
                        # stop were called from its other possible caller,
                        # the start of process_request when a new request comes
                        # in on the same token)
                        stop,
                        )
            else:
                # It'd be tempting to raise here, but typically being called
                # from a task, it wouldn't propagate any further either, and at
                # least here we have a logger.
                self.log.error("Requests shouldn't receive errors at the level of a TokenManager any more, but this did: %s", ev)
            if not ev.is_last:
                return True
        def on_end():
            if key in self.incoming_requests:
                # It may not be, especially if it was popped in `(pipe, stop) = self.incoming_requests.pop(keyu)` above
                # FIXME Decide if one of them is sufficient
                del self.incoming_requests[key]
            # no further cleanup to do here: any piggybackable ack was already flushed
            # out by the first response, and if there was not even a
            # NoResponse, something went wrong above (and we can't tell easily
            # here).
        stop = pipe.on_event(on_event)
        pipe.on_interest_end(on_end)

        self.incoming_requests[key] = (pipe, stop)

        self.context.render_to_pipe(pipe)

    def process_response(self, response):
        key = (response.token, response.remote)
        if key not in self.outgoing_requests:
            # maybe it was a multicast...
            key = (response.token, None)

        try:
            request = self.outgoing_requests[key]
        except KeyError:
            self.log.info("Response %r could not be matched to any request", response)
            return False
        else:
            self.log.debug("Response %r matched to request %r", response, request)

        # FIXME: there's a multicast aspect to that as well
        #
        # Is it necessary to look into .opt.observe here, wouldn't that better
        # be done by the higher-level code that knows about CoAP options?
        # Maybe, but at some point in TokenManager we *have* to look into the
        # options to see whether to expect a short- or long-running token.
        # Still, it would be an option not to send an is_last here and *always*
        # have the higher-level code indicate loss of interest in that exchange
        # when it detects that no more observations will follow.
        final = not (request.request.opt.observe == 0 and response.opt.observe is not None)

        if final:
            self.outgoing_requests.pop(key)

        request.add_response(response, is_last=final)
        return True

    #
    # implement RequestInterface
    #

    async def fill_or_recognize_remote(self, message):
        return await self.token_interface.fill_or_recognize_remote(message)

    def request(self, request):
        msg = request.request

        assert msg.code.is_request(), "Message code is not valid for request"

        # This might easily change, but right now, relying on the Context to
        # fill_remote early makes steps easier here.
        assert msg.remote is not None, "Remote not pre-populated"

        # FIXME: pick a suitably short one where available, and a longer one
        # for observations if many short ones are already in-flight
        msg.token = self.next_token()

        self.log.debug("Sending request - Token: %s, Remote: %s", msg.token.hex(), msg.remote)

        # A request sent over the multicast interface will only return a single
        # response and otherwise behave quite like an anycast request (which is
        # probably intended).
        if msg.remote.is_multicast:
            self.log.warning("Sending request to multicast via unicast request method")
            key = (msg.token, None)
        else:
            key = (msg.token, msg.remote)

        self.outgoing_requests[key] = request
        request.on_interest_end(functools.partial(self.outgoing_requests.pop, key, None))

        try:
            send_canceller = self.token_interface.send_message(msg, lambda: request.add_exception(error.MessageError))
        except Exception as e:
            request.add_exception(e)
            return

        if send_canceller is not None:
            # This needs to be called both when the requester cancels the
            # request, and when a response to the CON request comes in via a
            # different CON when the original ACK was lost, so the retransmits
            # can stop.
            #
            # FIXME: This might need a little sharper conditions: A fresh CON
            # should be sufficient to stop retransmits of a CON in a first
            # request, but when refreshing an observation, only an ACK tells us
            # that the updated observation got through. Also, multicast needs
            # to be an exception, but that generally needs handling here.
            #
            # It may be that it'd be wise to reduce the use of send_canceller
            # to situations when the request is actualy cancelled, and pass
            # some information to the token_interface about whether it should
            # keep an eye out for responses on that token and cancel
            # transmission accordingly.
            request.on_event(lambda ev: (send_canceller(), False)[1],
                    is_interest=False)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import asyncio

class TimeoutDict:
    """A dict-ish type whose entries live on a timeout; adding and accessing an
    item each refreshes the timeout.

    The timeout is a lower bound; items may live up to twice as long.

    The container is implemented incompletely, with additions made on demand.

    This is not thread safe.
    """

    def __init__(self, timeout: float):
        self.timeout = timeout
        """Timeout set on any access

        This can be changed at runtime, but changes only take effect """

        self._items = {}
        """The actual dictionary"""
        self._recently_accessed = None
        """Items accessed since the timeout last fired"""
        self._timeout = None
        """Canceler for the timeout function"""
        # Note: Without a __del__ implementation that even cancels, the object
        # will be kept alive by the main loop for a timeout

    def __getitem__(self, key):
        result = self._items[key]
        self._accessed(key)
        return result

    def __setitem__(self, key, value):
        self._items[key] = value
        self._accessed(key)

    def _start_over(self):
        """Clear _recently_accessed, set the timeout"""
        self._timeout = asyncio.get_running_loop().call_later(self.timeout, self._tick)
        self._recently_accessed = set()

    def _accessed(self, key):
        """Mark a key as recently accessed"""
        if self._timeout is None:
            self._start_over()
            # No need to add the key, it'll live for this duration anyway
        else:
            self._recently_accessed.add(key)

    def _tick(self):
        self._items = {k: v for (k, v) in self._items.items() if k in self._recently_accessed}
        if self._items:
            self._start_over()
        else:
            self._timeout = None
            self._recently_accessed = None



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This module contains the tools to send OSCORE secured messages.

It only deals with the algorithmic parts, the security context and protection
and unprotection of messages. It does not touch on the integration of OSCORE in
the larger aiocoap stack of having a context or requests; that's what
:mod:`aiocoap.transports.osore` is for.`"""

from __future__ import annotations

from collections import namedtuple
import io
import json
import binascii
import os
import os.path
import tempfile
import abc
from typing import Optional
import secrets
import warnings

from aiocoap.message import Message
from aiocoap.util import cryptography_additions, deprecation_getattr
from aiocoap.numbers import GET, POST, FETCH, CHANGED, UNAUTHORIZED, CONTENT
from aiocoap import error

from cryptography.hazmat.primitives.ciphers import aead
from cryptography.hazmat.primitives.kdf.hkdf import HKDF
from cryptography.hazmat.primitives import hashes
import cryptography.hazmat.backends
import cryptography.exceptions
from cryptography.hazmat.primitives import asymmetric, serialization
from cryptography.hazmat.primitives.asymmetric.utils import decode_dss_signature, encode_dss_signature

import cbor2 as cbor

import filelock

MAX_SEQNO = 2**40 - 1

# Relevant values from the IANA registry "CBOR Object Signing and Encryption (COSE)"
COSE_KID = 4
COSE_PIV = 6
COSE_KID_CONTEXT = 10
# from RFC9338
COSE_COUNTERSIGNATURE0 = 12
# from draft-ietf-lake-edhoc-19, guessing the value
COSE_KCCS = 13131313

COMPRESSION_BITS_N = 0b111
COMPRESSION_BIT_K = 0b1000
COMPRESSION_BIT_H = 0b10000
COMPRESSION_BIT_G = 0b100000 # Group Flag from draft-ietf-core-oscore-groupcomm-10
COMPRESSION_BITS_RESERVED = 0b11000000

# While the original values were simple enough to be used in literals, starting
# with oscore-groupcomm we're using more compact values

INFO_TYPE_KEYSTREAM_REQUEST = True
INFO_TYPE_KEYSTREAM_RESPONSE = False

class CodeStyle(namedtuple("_CodeStyle", ("request", "response"))):
    @classmethod
    def from_request(cls, request) -> CodeStyle:
        if request == FETCH:
            return cls.FETCH_CONTENT
        elif request == POST:
            return cls.POST_CHANGED
        else:
            raise ValueError("Invalid request code %r" % request)

CodeStyle.FETCH_CONTENT = CodeStyle(FETCH, CONTENT)
CodeStyle.POST_CHANGED = CodeStyle(POST, CHANGED)


class DeterministicKey:
    """Singleton to indicate that for this key member no public or private key
    is available because it is the Deterministic Client (see
    <https://www.ietf.org/archive/id/draft-amsuess-core-cachable-oscore-01.html>)

    This is highly experimental not only from an implementation but also from a
    specification point of view. The specification has not received adaequate
    review that would justify using it in any non-experimental scenario.
    """
DETERMINISTIC_KEY = DeterministicKey()
del DeterministicKey

class NotAProtectedMessage(error.Error, ValueError):
    """Raised when verification is attempted on a non-OSCORE message"""

    def __init__(self, message, plain_message):
        super().__init__(message)
        self.plain_message = plain_message

class ProtectionInvalid(error.Error, ValueError):
    """Raised when verification of an OSCORE message fails"""

class DecodeError(ProtectionInvalid):
    """Raised when verification of an OSCORE message fails because CBOR or compressed data were erroneous"""

class ReplayError(ProtectionInvalid):
    """Raised when verification of an OSCORE message fails because the sequence numbers was already used"""

class ReplayErrorWithEcho(ProtectionInvalid, error.RenderableError):
    """Raised when verification of an OSCORE message fails because the
    recipient replay window is uninitialized, but a 4.01 Echo can be
    constructed with the data in the exception that can lead to the client
    assisting in replay window recovery"""
    def __init__(self, secctx, request_id, echo):
        self.secctx = secctx
        self.request_id = request_id
        self.echo = echo

    def to_message(self):
        inner = Message(
                code=UNAUTHORIZED,
                echo=self.echo,
                )
        outer, _ = self.secctx.protect(inner, request_id=self.request_id)
        return outer

class ContextUnavailable(error.Error, ValueError):
    """Raised when a context is (currently or permanently) unavailable for
    protecting or unprotecting a message"""

class RequestIdentifiers:
    """A container for details that need to be passed along from the
    (un)protection of a request to the (un)protection of the response; these
    data ensure that the request-response binding process works by passing
    around the request's partial IV.

    Users of this module should never create or interact with instances, but
    just pass them around.
    """
    def __init__(self, kid, partial_iv, nonce, can_reuse_nonce, request_code):
        self.kid = kid
        self.partial_iv = partial_iv
        self.nonce = nonce
        self.can_reuse_nonce = can_reuse_nonce
        self.code_style = CodeStyle.from_request(request_code)

        self.request_hash = None

    def get_reusable_nonce_and_piv(self):
        """Return the nonce and the partial IV if can_reuse_nonce is True, and
        set can_reuse_nonce to False."""

        if self.can_reuse_nonce:
            self.can_reuse_nonce = False
            return (self.nonce, self.partial_iv)
        else:
            return (None, None)

def _xor_bytes(a, b):
    assert len(a) == len(b), "XOR needs consistent lengths"
    # FIXME is this an efficient thing to do, or should we store everything
    # that possibly needs xor'ing as long integers with an associated length?
    return bytes(_a ^ _b for (_a, _b) in zip(a, b))

class AeadAlgorithm(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def encrypt(cls, plaintext, aad, key, iv):
        """Return ciphertext + tag for given input data"""

    @abc.abstractmethod
    def decrypt(cls, ciphertext_and_tag, aad, key, iv):
        """Reverse encryption. Must raise ProtectionInvalid on any error
        stemming from untrusted data."""

    @staticmethod
    def _build_encrypt0_structure(protected, external_aad):
        assert protected == {}, "Unexpected data in protected bucket"
        protected_serialized = b'' # were it into an empty dict, it'd be the cbor dump
        enc_structure = ['Encrypt0', protected_serialized, external_aad]

        return cbor.dumps(enc_structure)

class AES_CCM(AeadAlgorithm, metaclass=abc.ABCMeta):
    """AES-CCM implemented using the Python cryptography library"""

    @classmethod
    def encrypt(cls, plaintext, aad, key, iv):
        return aead.AESCCM(key, cls.tag_bytes).encrypt(iv, plaintext, aad)

    @classmethod
    def decrypt(cls, ciphertext_and_tag, aad, key, iv):
        try:
            return aead.AESCCM(key, cls.tag_bytes).decrypt(iv, ciphertext_and_tag, aad)
        except cryptography.exceptions.InvalidTag:
            raise ProtectionInvalid("Tag invalid")

class AES_CCM_16_64_128(AES_CCM):
    # from RFC8152 and draft-ietf-core-object-security-0[012] 3.2.1
    value = 10
    key_bytes = 16 # 128-bit key
    tag_bytes = 8 # 64-bit tag
    iv_bytes = 13 # 13-byte nonce

class AES_CCM_16_64_256(AES_CCM):
    # from RFC8152
    value = 11
    key_bytes = 32 # 256-bit key
    tag_bytes = 8 # 64-bit tag
    iv_bytes = 13 # 13-byte nonce

class AES_CCM_64_64_128(AES_CCM):
    # from RFC8152
    value = 12
    key_bytes = 16 # 128-bit key
    tag_bytes = 8 # 64-bit tag
    iv_bytes = 7 # 7-byte nonce

class AES_CCM_64_64_256(AES_CCM):
    # from RFC8152
    value = 13
    key_bytes = 32 # 256-bit key
    tag_bytes = 8 # 64-bit tag
    iv_bytes = 7 # 7-byte nonce

class AES_CCM_16_128_128(AES_CCM):
    # from RFC8152
    value = 30
    key_bytes = 16 # 128-bit key
    tag_bytes = 16 # 128-bit tag
    iv_bytes = 13 # 13-byte nonce

class AES_CCM_16_128_256(AES_CCM):
    # from RFC8152
    value = 31
    key_bytes = 32 # 256-bit key
    tag_bytes = 16 # 128-bit tag
    iv_bytes = 13 # 13-byte nonce

class AES_CCM_64_128_128(AES_CCM):
    # from RFC8152
    value = 32
    key_bytes = 16 # 128-bit key
    tag_bytes = 16 # 128-bit tag
    iv_bytes = 7 # 7-byte nonce

class AES_CCM_64_128_256(AES_CCM):
    # from RFC8152
    value = 33
    key_bytes = 32 # 256-bit key
    tag_bytes = 16 # 128-bit tag
    iv_bytes = 7 # 7-byte nonce


class AES_GCM(AeadAlgorithm, metaclass=abc.ABCMeta):
    """AES-GCM implemented using the Python cryptography library"""

    iv_bytes = 12 # 96 bits fixed size of the nonce

    @classmethod
    def encrypt(cls, plaintext, aad, key, iv):
        return aead.AESGCM(key).encrypt(iv, plaintext, aad)

    @classmethod
    def decrypt(cls, ciphertext_and_tag, aad, key, iv):
        try:
            return aead.AESGCM(key).decrypt(iv, ciphertext_and_tag, aad)
        except cryptography.exceptions.InvalidTag:
            raise ProtectionInvalid("Tag invalid")

class A128GCM(AES_GCM):
    # from RFC8152
    value = 1
    key_bytes = 16 # 128-bit key
    tag_bytes = 16 # 128-bit tag

class A192GCM(AES_GCM):
    # from RFC8152
    value = 2
    key_bytes = 24 # 192-bit key
    tag_bytes = 16 # 128-bit tag

class A256GCM(AES_GCM):
    # from RFC8152
    value = 3
    key_bytes = 32 # 256-bit key
    tag_bytes = 16 # 128-bit tag

class ChaCha20Poly1305(AeadAlgorithm):
    # from RFC8152
    value = 24
    key_bytes = 32 # 256-bit key
    tag_bytes = 16 # 128-bit tag
    iv_bytes = 12 # 96-bit nonce

    @classmethod
    def encrypt(cls, plaintext, aad, key, iv):
        return aead.ChaCha20Poly1305(key).encrypt(iv, plaintext, aad)

    @classmethod
    def decrypt(cls, ciphertext_and_tag, aad, key, iv):
        try:
            return aead.ChaCha20Poly1305(key).decrypt(iv, ciphertext_and_tag, aad)
        except cryptography.exceptions.InvalidTag:
            raise ProtectionInvalid("Tag invalid")

class AlgorithmCountersign(metaclass=abc.ABCMeta):
    """A fully parameterized COSE countersign algorithm

    An instance is able to provide all the alg_signature, par_countersign and
    par_countersign_key parameters taht go into the Group OSCORE algorithms
    field.
    """
    @abc.abstractmethod
    def sign(self, body, external_aad, private_key):
        """Return the signature produced by the key when using
        CounterSignature0 as describe in draft-ietf-cose-countersign-01"""

    @abc.abstractmethod
    def verify(self, signature, body, external_aad, public_key):
        """Verify a signature in analogy to sign"""

    @abc.abstractmethod
    def generate(self):
        """Return a usable private key"""

    @abc.abstractmethod
    def public_from_private(self, private_key):
        """Given a private key, derive the publishable key"""

    @staticmethod
    def _build_countersign_structure(body, external_aad):
        countersign_structure = [
                "CounterSignature0",
                b"",
                b"",
                external_aad,
                body
                ]
        tobesigned = cbor.dumps(countersign_structure)
        return tobesigned

    @property
    @abc.abstractproperty
    def signature_length(self):
        """The length of a signature using this algorithm"""

    @property
    @abc.abstractproperty
    def curve_number(self):
        """Registered curve number used with this algorithm.

        Only used for verification of credentials' details"""

class AlgorithmStaticStatic(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def staticstatic(self, private_key, public_key):
        """Derive a shared static-static secret from a private and a public key"""

class Ed25519(AlgorithmCountersign):
    def sign(self, body, aad, private_key):
        private_key = asymmetric.ed25519.Ed25519PrivateKey.from_private_bytes(private_key)
        return private_key.sign(self._build_countersign_structure(body, aad))

    def verify(self, signature, body, aad, public_key):
        public_key = asymmetric.ed25519.Ed25519PublicKey.from_public_bytes(public_key)
        try:
            public_key.verify(signature, self._build_countersign_structure(body, aad))
        except cryptography.exceptions.InvalidSignature:
            raise ProtectionInvalid("Signature mismatch")

    def generate(self):
        key = asymmetric.ed25519.Ed25519PrivateKey.generate()
        # FIXME: We could avoid handing the easy-to-misuse bytes around if the
        # current algorithm interfaces did not insist on passing the
        # exchangable representations -- and generally that should be more
        # efficient.
        return key.private_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PrivateFormat.Raw,
            encryption_algorithm=serialization.NoEncryption(),
            )

    def public_from_private(self, private_key):
        private_key = asymmetric.ed25519.Ed25519PrivateKey.from_private_bytes(private_key)
        public_key = private_key.public_key()
        return public_key.public_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PublicFormat.Raw,
            )

    value = -8
    curve_number = 6

    signature_length = 64

class EcdhSsHkdf256(AlgorithmStaticStatic):
    # FIXME: This class uses the Edwards keys as private and public keys, and
    # not the converted ones. This will be problematic if pairwise-only
    # contexts are to be set up.

    value = -27 # FIXME: or -28? (see shepherd review)

    # FIXME these two will be different when using the Montgomery keys directly

    # This one will only be used when establishing and distributing pairwise-only keys
    generate = Ed25519.generate
    public_from_private = Ed25519.public_from_private

    def staticstatic(self, private_key, public_key):
        private_key = asymmetric.ed25519.Ed25519PrivateKey.from_private_bytes(private_key)
        private_key = cryptography_additions.sk_to_curve25519(private_key)

        public_key = asymmetric.ed25519.Ed25519PublicKey.from_public_bytes(public_key)
        public_key = cryptography_additions.pk_to_curve25519(public_key)

        return private_key.exchange(public_key)

class ECDSA_SHA256_P256(AlgorithmCountersign, AlgorithmStaticStatic):
    # Trying a new construction approach -- should work just as well given
    # we're just passing Python objects around
    def from_public_parts(self, x: bytes, y: bytes):
        """Create a public key from its COSE values"""
        return asymmetric.ec.EllipticCurvePublicNumbers(
                int.from_bytes(x, 'big'),
                int.from_bytes(y, 'big'),
                asymmetric.ec.SECP256R1()
                ).public_key()

    def from_private_parts(self, x: bytes, y: bytes, d: bytes):
        public_numbers = self.from_public_parts(x, y).public_numbers()
        private_numbers = asymmetric.ec.EllipticCurvePrivateNumbers(
                int.from_bytes(d, 'big'),
                public_numbers)
        return private_numbers.private_key()

    def sign(self, body, aad, private_key):
        der_signature = private_key.sign(self._build_countersign_structure(body, aad), asymmetric.ec.ECDSA(hashes.SHA256()))
        (r, s) = decode_dss_signature(der_signature)

        return r.to_bytes(32, "big") + s.to_bytes(32, "big")

    def verify(self, signature, body, aad, public_key):
        r = signature[:32]
        s = signature[32:]
        r = int.from_bytes(r, "big")
        s = int.from_bytes(s, "big")
        der_signature = encode_dss_signature(r, s)
        try:
            public_key.verify(der_signature, self._build_countersign_structure(body, aad), asymmetric.ec.ECDSA(hashes.SHA256()))
        except cryptography.exceptions.InvalidSignature:
            raise ProtectionInvalid("Signature mismatch")

    def generate(self):
        return asymmetric.ec.generate_private_key(asymmetric.ec.SECP256R1())

    def public_from_private(self, private_key):
        return private_key.public_key()

    def staticstatic(self, private_key, public_key):
        return private_key.exchange(asymmetric.ec.ECDH(), public_key)

    value = -7 # FIXME: when used as a static-static algorithm, does this become -27? see shepherd review.
    curve_number = 1

    signature_length = 64

algorithms = {
        'AES-CCM-16-64-128': AES_CCM_16_64_128(),
        'AES-CCM-16-64-256': AES_CCM_16_64_256(),
        'AES-CCM-64-64-128': AES_CCM_64_64_128(),
        'AES-CCM-64-64-256': AES_CCM_64_64_256(),
        'AES-CCM-16-128-128': AES_CCM_16_128_128(),
        'AES-CCM-16-128-256': AES_CCM_16_128_256(),
        'AES-CCM-64-128-128': AES_CCM_64_128_128(),
        'AES-CCM-64-128-256': AES_CCM_64_128_256(),
        'ChaCha20/Poly1305': ChaCha20Poly1305(),
        'A128GCM': A128GCM(),
        'A192GCM': A192GCM(),
        'A256GCM': A256GCM(),
        }

# algorithms with full parameter set
algorithms_countersign = {
        # maybe needs a different name...
        'EdDSA on Ed25519': Ed25519(),
        'ECDSA w/ SHA-256 on P-256': ECDSA_SHA256_P256(),
        }

algorithms_staticstatic = {
        'ECDH-SS + HKDF-256': EcdhSsHkdf256(),
        }

DEFAULT_ALGORITHM = 'AES-CCM-16-64-128'

_hash_backend = cryptography.hazmat.backends.default_backend()
hashfunctions = {
        'sha256': hashes.SHA256(),
        'sha384': hashes.SHA384(),
        'sha512': hashes.SHA512(),
        }

DEFAULT_HASHFUNCTION = 'sha256'

DEFAULT_WINDOWSIZE = 32

class BaseSecurityContext:
    # The protection and unprotection functions will use the Group OSCORE AADs
    # rather than the regular OSCORE AADs. (Ie. alg_signature_enc etc are added to
    # the algorithms, and request_kid_context, OSCORE_option, sender_auth_cred and
    # gm_cred are added).
    #
    # This is not necessarily identical to is_signing (as pairwise contexts use
    # this but don't sign), and is distinct from the added OSCORE option in the
    # AAD (as that's only applicable for the external AAD as extracted for
    # signing and signature verification purposes).
    external_aad_is_group = False

    # Authentication information carried with this security context; managed
    # externally by whatever creates the security context.
    authenticated_claims = []

    # AEAD algorithm. This may only be None in group contexts that do not use pairwise mode.
    alg_aead: Optional[AeadAlgorithm]

    @property
    def algorithm(self):
        warnings.warn("Property was renamed to 'alg_aead'", DeprecationWarning, stacklevel=2)
        return self.alg_aead
    @algorithm.setter
    def algorithm(self, value):
        warnings.warn("Property was renamed to 'alg_aead'", DeprecationWarning, stacklevel=2)
        self.alg_aead = value

    def _construct_nonce(self, partial_iv_short, piv_generator_id):
        pad_piv = b"\0" * (5 - len(partial_iv_short))

        s = bytes([len(piv_generator_id)])
        pad_id = b'\0' * (self.alg_aead.iv_bytes - 6 - len(piv_generator_id))

        components = s + \
                pad_id + \
                piv_generator_id + \
                pad_piv + \
                partial_iv_short

        nonce = _xor_bytes(self.common_iv, components)

        return nonce

    def _extract_external_aad(self, message, request_id, local_is_sender: bool) -> bytes:
        """Build the serialized external AAD from information in the message
        and the request_id.

        Information about whether the local context is the sender of the
        message is only relevant to group contexts, where it influences whose
        authentication credentials are placed in the AAD.
        """
        # If any option were actually Class I, it would be something like
        #
        # the_options = pick some of(message)
        # class_i_options = Message(the_options).opt.encode()

        oscore_version = 1
        class_i_options = b""
        if request_id.request_hash is not None:
            class_i_options = Message(request_hash=request_id.request_hash).opt.encode()

        algorithms = [self.alg_aead.value]
        if self.external_aad_is_group:
            algorithms.append(self.alg_signature_enc.value)
            algorithms.append(self.alg_signature.value)
            algorithms.append(self.alg_pairwise_key_agreement.value)

        external_aad = [
                oscore_version,
                algorithms,
                request_id.kid,
                request_id.partial_iv,
                class_i_options,
                ]

        if self.external_aad_is_group:
            # FIXME: We may need to carry this over in the request_id when
            # observation span group rekeyings
            external_aad.append(self.id_context)

            assert message.opt.object_security is not None, "Double OSCORE"
            external_aad.append(message.opt.object_security)

            if local_is_sender:
                external_aad.append(self.sender_auth_cred)
            else:
                external_aad.append(self.recipient_auth_cred)
            external_aad.append(self.group_manager_cred)

        external_aad = cbor.dumps(external_aad)

        return external_aad

# FIXME pull interface components from SecurityContext up here
class CanProtect(BaseSecurityContext, metaclass=abc.ABCMeta):
    # The protection function will add a signature acccording to the context's
    # alg_signature attribute if this is true
    is_signing = False

    # Send the KID when protecting responses
    #
    # Once group pairwise mode is implemented, this will need to become a
    # parameter to protect(), which is stored at the point where the incoming
    # context is turned into an outgoing context. (Currently, such a mechanism
    # isn't there yet, and oscore_wrapper protects responses with the very same
    # context they came in on).
    responses_send_kid = False

    @staticmethod
    def _compress(protected, unprotected, ciphertext):
        """Pack the untagged COSE_Encrypt0 object described by the *args
        into two bytestrings suitable for the Object-Security option and the
        message body"""

        if protected:
            raise RuntimeError("Protection produced a message that has uncompressable fields.")

        piv = unprotected.pop(COSE_PIV, b"")
        if len(piv) > COMPRESSION_BITS_N:
            raise ValueError("Can't encode overly long partial IV")

        firstbyte = len(piv)
        if COSE_KID in unprotected:
            firstbyte |= COMPRESSION_BIT_K
            kid_data = unprotected.pop(COSE_KID)
        else:
            kid_data = b""

        if COSE_KID_CONTEXT in unprotected:
            firstbyte |= COMPRESSION_BIT_H
            kid_context = unprotected.pop(COSE_KID_CONTEXT)
            s = len(kid_context)
            if s > 255:
                raise ValueError("KID Context too long")
            s_kid_context = bytes((s,)) + kid_context
        else:
            s_kid_context = b""

        if COSE_COUNTERSIGNATURE0 in unprotected:
            firstbyte |= COMPRESSION_BIT_G

            # In theory at least. In practice, that's an empty value to later
            # be squished in when the compressed option value is available for
            # signing.
            ciphertext += unprotected.pop(COSE_COUNTERSIGNATURE0)

        if unprotected:
            raise RuntimeError("Protection produced a message that has uncompressable fields.")

        if firstbyte:
            option = bytes([firstbyte]) + piv + s_kid_context + kid_data
        else:
            option = b""

        return (option, ciphertext)

    def protect(self, message, request_id=None, *, kid_context=True):
        """Given a plain CoAP message, create a protected message that contains
        message's options in the inner or outer CoAP message as described in
        OSCOAP.

        If the message is a response to a previous message, the additional data
        from unprotecting the request are passed in as request_id. When
        request data is present, its partial IV is reused if possible. The
        security context's ID context is encoded in the resulting message
        unless kid_context is explicitly set to a False; other values for the
        kid_context can be passed in as byte string in the same parameter.
        """

        assert (request_id is None) == message.code.is_request(), "Requestishness of code to protect does not match presence of request ID"

        outer_message, plaintext = self._split_message(message, request_id)

        protected = {}
        nonce = None
        unprotected = {}
        if request_id is not None:
            nonce, partial_iv_short = request_id.get_reusable_nonce_and_piv()
            if nonce is not None:
                partial_iv_generated_by = request_id.kid

        if nonce is None:
            nonce, partial_iv_short = self._build_new_nonce()
            partial_iv_generated_by = self.sender_id

            unprotected[COSE_PIV] = partial_iv_short

        if message.code.is_request():
            unprotected[COSE_KID] = self.sender_id

            request_id = RequestIdentifiers(self.sender_id, partial_iv_short, nonce, can_reuse_nonce=None, request_code=outer_message.code)

            if kid_context is True:
                if self.id_context is not None:
                    unprotected[COSE_KID_CONTEXT] = self.id_context
            elif kid_context is not False:
                unprotected[COSE_KID_CONTEXT] = kid_context
        else:
            if self.responses_send_kid:
                unprotected[COSE_KID] = self.sender_id

        # Putting in a dummy value as the signature calculation will already need some of the compression result
        if self.is_signing:
            unprotected[COSE_COUNTERSIGNATURE0] = b""
        # FIXME: Running this twice quite needlessly (just to get the object_security option for sending)
        option_data, _ = self._compress(protected, unprotected, b"")

        outer_message.opt.object_security = option_data

        external_aad = self._extract_external_aad(outer_message, request_id, local_is_sender=True)

        aad = self.alg_aead._build_encrypt0_structure(protected, external_aad)

        key = self._get_sender_key(outer_message, external_aad, plaintext, request_id)

        ciphertext = self.alg_aead.encrypt(plaintext, aad, key, nonce)

        _, payload = self._compress(protected, unprotected, ciphertext)

        if self.is_signing:
            signature = self.alg_signature.sign(payload, external_aad, self.private_key)
            keystream = self._kdf_for_keystreams(
                    partial_iv_generated_by,
                    partial_iv_short,
                    self.group_encryption_key,
                    self.sender_id,
                    INFO_TYPE_KEYSTREAM_REQUEST if message.code.is_request() else INFO_TYPE_KEYSTREAM_RESPONSE,
                    )
            encrypted_signature = _xor_bytes(signature, keystream)
            payload += encrypted_signature
        outer_message.payload = payload

        # FIXME go through options section

        # the request_id in the second argument should be discarded by the
        # caller when protecting a response -- is that reason enough for an
        # `if` and returning None?
        return outer_message, request_id

    def _get_sender_key(self, outer_message, aad, plaintext, request_id):
        """Customization hook of the protect function

        While most security contexts have a fixed sender key, deterministic
        requests need to shake up a few things. They need to modify the outer
        message, as well as the request_id as it will later be used to
        unprotect the response."""
        return self.sender_key

    def _split_message(self, message, request_id):
        """Given a protected message, return the outer message that contains
        all Class I and Class U options (but without payload or Object-Security
        option), and the encoded inner message that contains all Class E
        options and the payload.

        This leaves the messages' remotes unset."""

        if message.code.is_request():
            outer_host = message.opt.uri_host
            proxy_uri = message.opt.proxy_uri

            inner_message = message.copy(
                    uri_host=None,
                    uri_port=None,
                    proxy_uri=None,
                    proxy_scheme=None,
                    )
            inner_message.remote = None

            if proxy_uri is not None:
                # Use set_request_uri to split up the proxy URI into its
                # components; extract, preserve and clear them.
                inner_message.set_request_uri(proxy_uri, set_uri_host=False)
                if inner_message.opt.proxy_uri is not None:
                    raise ValueError("Can not split Proxy-URI into options")
                outer_uri = inner_message.remote.uri_base
                inner_message.remote = None
                inner_message.opt.proxy_scheme = None

            if message.opt.observe is None:
                outer_code = POST
            else:
                outer_code = FETCH
        else:
            outer_host = None
            proxy_uri = None

            inner_message = message.copy()

            outer_code = request_id.code_style.response

        # no max-age because these are always successsful responses
        outer_message = Message(code=outer_code,
                uri_host=outer_host,
                observe=None if message.code.is_response() else message.opt.observe,
                )
        if proxy_uri is not None:
            outer_message.set_request_uri(outer_uri)

        plaintext = bytes([inner_message.code]) + inner_message.opt.encode()
        if inner_message.payload:
            plaintext += bytes([0xFF])
            plaintext += inner_message.payload

        return outer_message, plaintext

    def _build_new_nonce(self):
        """This implements generation of a new nonce, assembled as per Figure 5
        of draft-ietf-core-object-security-06. Returns the shortened partial IV
        as well."""
        seqno = self.new_sequence_number()

        partial_iv = seqno.to_bytes(5, 'big')

        return (self._construct_nonce(partial_iv, self.sender_id), partial_iv.lstrip(b'\0') or b'\0')

    # sequence number handling

    def new_sequence_number(self):
        """Return a new sequence number; the implementation is responsible for
        never returning the same value twice in a given security context.

        May raise ContextUnavailable."""
        retval = self.sender_sequence_number
        if retval >= MAX_SEQNO:
            raise ContextUnavailable("Sequence number too large, context is exhausted.")
        self.sender_sequence_number += 1
        self.post_seqnoincrease()
        return retval

    # implementation defined

    @abc.abstractmethod
    def post_seqnoincrease(self):
        """Ensure that sender_sequence_number is stored"""
        raise

    def context_from_response(self, unprotected_bag) -> CanUnprotect:
        """When receiving a response to a request protected with this security
        context, pick the security context with which to unprotect the response
        given the unprotected information from the Object-Security option.

        This allow picking the right security context in a group response, and
        helps getting a new short-lived context for B.2 mode. The default
        behaivor is returning self.
        """
        return self # FIXME justify by moving into a mixin for CanProtectAndUnprotect

class CanUnprotect(BaseSecurityContext):
    def unprotect(self, protected_message, request_id=None):
        assert (request_id is not None) == protected_message.code.is_response(), "Requestishness of code to unprotect does not match presence of request ID"
        is_response = protected_message.code.is_response()

        # Set to a raisable exception on replay check failures; it will be
        # raised, but the package may still be processed in the course of Echo handling.
        replay_error = None

        protected_serialized, protected, unprotected, ciphertext = self._extract_encrypted0(protected_message)

        if protected:
            raise ProtectionInvalid("The protected field is not empty")

        # FIXME check for duplicate keys in protected

        if unprotected.pop(COSE_KID_CONTEXT, self.id_context) != self.id_context:
            # FIXME is this necessary?
            raise ProtectionInvalid("Sender ID context does not match")

        if unprotected.pop(COSE_KID, self.recipient_id) != self.recipient_id:
            # for most cases, this is caught by the session ID dispatch, but in
            # responses (where explicit sender IDs are atypical), this is a
            # valid check
            raise ProtectionInvalid("Sender ID does not match")

        if COSE_PIV not in unprotected:
            if not is_response:
                raise ProtectionInvalid("No sequence number provided in request")

            nonce = request_id.nonce
            seqno = None # sentinel for not striking out anyting
            partial_iv_short = request_id.partial_iv
            partial_iv_generated_by = request_id.kid
        else:
            partial_iv_short = unprotected.pop(COSE_PIV)
            partial_iv_generated_by = self.recipient_id

            nonce = self._construct_nonce(partial_iv_short, self.recipient_id)

            seqno = int.from_bytes(partial_iv_short, 'big')

            if not is_response:
                if not self.recipient_replay_window.is_initialized():
                    replay_error = ReplayError("Sequence number check unavailable")
                elif not self.recipient_replay_window.is_valid(seqno):
                    replay_error = ReplayError("Sequence number was re-used")

                if replay_error is not None and self.echo_recovery is None:
                    # Don't even try decoding if there is no reason to
                    raise replay_error

                request_id = RequestIdentifiers(self.recipient_id, partial_iv_short, nonce, can_reuse_nonce=replay_error is None, request_code=protected_message.code)

        if unprotected.pop(COSE_COUNTERSIGNATURE0, None) is not None:
            try:
                alg_signature = self.alg_signature
            except NameError:
                raise DecodeError("Group messages can not be decoded with this non-group context")

            siglen = alg_signature.signature_length
            if len(ciphertext) < siglen:
                raise DecodeError("Message too short for signature")
            encrypted_signature = ciphertext[-siglen:]

            keystream = self._kdf_for_keystreams(
                    partial_iv_generated_by,
                    partial_iv_short,
                    self.group_encryption_key,
                    self.recipient_id,
                    INFO_TYPE_KEYSTREAM_REQUEST if protected_message.code.is_request() else INFO_TYPE_KEYSTREAM_RESPONSE,
                    )
            signature = _xor_bytes(encrypted_signature, keystream)

            ciphertext = ciphertext[:-siglen]
        else:
            signature = None

        if unprotected:
            raise DecodeError("Unsupported unprotected option")

        if len(ciphertext) < self.alg_aead.tag_bytes + 1: # +1 assures access to plaintext[0] (the code)
            raise ProtectionInvalid("Ciphertext too short")

        external_aad = self._extract_external_aad(protected_message, request_id, local_is_sender=False)
        enc_structure = ['Encrypt0', protected_serialized, external_aad]
        aad = cbor.dumps(enc_structure)

        key = self._get_recipient_key(protected_message)

        plaintext = self.alg_aead.decrypt(ciphertext, aad, key, nonce)

        self._post_decrypt_checks(external_aad, plaintext, protected_message, request_id)

        if not is_response and seqno is not None and replay_error is None:
            self.recipient_replay_window.strike_out(seqno)

        if signature is not None:
            # Only doing the expensive signature validation once the cheaper decyrption passed
            alg_signature.verify(signature, ciphertext, external_aad, self.recipient_public_key)

        # FIXME add options from unprotected

        unprotected_message = Message(code=plaintext[0])
        unprotected_message.payload = unprotected_message.opt.decode(plaintext[1:])

        try_initialize = not self.recipient_replay_window.is_initialized() and \
                self.echo_recovery is not None
        if try_initialize:
            if protected_message.code.is_request():
                # Either accept into replay window and clear replay error, or raise
                # something that can turn into a 4.01,Echo response
                if unprotected_message.opt.echo == self.echo_recovery:
                    self.recipient_replay_window.initialize_from_freshlyseen(seqno)
                    replay_error = None
                else:
                    raise ReplayErrorWithEcho(secctx=self, request_id=request_id, echo=self.echo_recovery)
            else:
                # We can initialize the replay window from a response as well.
                # The response is guaranteed fresh as it was AEAD-decoded to
                # match a request sent by this process.
                #
                # This is rare, as it only works when the server uses an own
                # sequence number, eg. when sending a notification or when
                # acting again on a retransmitted safe request whose response
                # it did not cache.
                #
                # Nothing bad happens if we can't make progress -- we just
                # don't initialize the replay window that wouldn't have been
                # checked for a response anyway.
                if seqno is not None:
                    self.recipient_replay_window.initialize_from_freshlyseen(seqno)

        if replay_error is not None:
            raise replay_error

        if unprotected_message.code.is_request():
            if protected_message.opt.observe != 0:
                unprotected_message.opt.observe = None
        else:
            if protected_message.opt.observe is not None:
                # -1 ensures that they sort correctly in later reordering
                # detection. Note that neither -1 nor high (>3 byte) sequence
                # numbers can be serialized in the Observe option, but they are
                # in this implementation accepted for passing around.
                unprotected_message.opt.observe = -1 if seqno is None else seqno

        return unprotected_message, request_id

    def _get_recipient_key(self, protected_message):
        """Customization hook of the unprotect function

        While most security contexts have a fixed recipient key, deterministic
        requests build it on demand."""
        return self.recipient_key

    def _post_decrypt_checks(self, aad, plaintext, protected_message, request_id):
        """Customization hook of the unprotect function after decryption

        While most security contexts are good with the default checks,
        deterministic requests need to perform additional checks while AAD and
        plaintext information is still available, and modify the request_id for
        the later protection step of the response."""

    @staticmethod
    def _uncompress(option_data, payload):
        if option_data == b"":
            firstbyte = 0
        else:
            firstbyte = option_data[0]
            tail = option_data[1:]

        unprotected = {}

        if firstbyte & COMPRESSION_BITS_RESERVED:
            raise DecodeError("Protected data uses reserved fields")

        pivsz = firstbyte & COMPRESSION_BITS_N
        if pivsz:
            if len(tail) < pivsz:
                raise DecodeError("Partial IV announced but not present")
            unprotected[COSE_PIV] = tail[:pivsz]
            tail = tail[pivsz:]

        if firstbyte & COMPRESSION_BIT_H:
            # kid context hint
            s = tail[0]
            if len(tail) - 1 < s:
                raise DecodeError("Context hint announced but not present")
            tail = tail[1:]
            unprotected[COSE_KID_CONTEXT] = tail[:s]
            tail = tail[s:]

        if firstbyte & COMPRESSION_BIT_K:
            kid = tail
            unprotected[COSE_KID] = kid

        if firstbyte & COMPRESSION_BIT_G:
            # Not really; As this is (also) used early on (before the KID
            # context is even known, because it's just getting extracted), this
            # is returning an incomplete value here and leaves it to the later
            # processing to strip the right number of bytes from the ciphertext
            unprotected[COSE_COUNTERSIGNATURE0] = b""

        return b"", {}, unprotected, payload

    @classmethod
    def _extract_encrypted0(cls, message):
        if message.opt.object_security is None:
            raise NotAProtectedMessage("No Object-Security option present", message)

        protected_serialized, protected, unprotected, ciphertext = cls._uncompress(message.opt.object_security, message.payload)
        return protected_serialized, protected, unprotected, ciphertext

    # implementation defined

    def context_for_response(self) -> CanProtect:
        """After processing a request with this context, with which security
        context should an outgoing response be protected? By default, it's the
        same context."""
        # FIXME: Is there any way in which the handler may want to influence
        # the decision taken here? Or would, then, the handler just call a more
        # elaborate but similar function when setting the response's remote
        # already?
        return self # FIXME justify by moving into a mixin for CanProtectAndUnprotect

class SecurityContextUtils(BaseSecurityContext):
    def _kdf(self, salt, ikm, role_id, out_type):
        """The HKDF as used to derive sender and recipient key and IV in
        RFC8613 Section 3.2.1, and analogously the Group Encryption Key of oscore-groupcomm.
        """
        if out_type == 'Key':
            out_bytes = self.alg_aead.key_bytes
        elif out_type == 'IV':
            out_bytes = self.alg_aead.iv_bytes
        elif out_type == "Group Encryption Key":
            out_bytes = self.alg_signature_enc.key_bytes
        else:
            raise ValueError("Output type not recognized")

        info = [
            role_id,
            self.id_context,
            self.alg_aead.value,
            out_type,
            out_bytes
            ]
        return self._kdf_lowlevel(salt, ikm, info, out_bytes)

    def _kdf_for_keystreams(self, piv_generated_by, salt, ikm, role_id, out_type):
        """The HKDF as used to derive the keystreams of oscore-groupcomm."""

        out_bytes = self.alg_signature.signature_length

        assert out_type in (INFO_TYPE_KEYSTREAM_REQUEST, INFO_TYPE_KEYSTREAM_RESPONSE), "Output type not recognized"

        info = [
            piv_generated_by,
            self.id_context,
            out_type,
            out_bytes
            ]
        return self._kdf_lowlevel(salt, ikm, info, out_bytes)

    def _kdf_lowlevel(self, salt: bytes, ikm: bytes, info: list, l: int) -> bytes:
        """The HKDF function as used in RFC8613 and oscore-groupcomm (notated
        there as ``something = HKDF(...)``

        Note that `info` typically contains `L` at some point.

        When `info` takes the conventional structure of pid, id_context,
        ald_aead, type, L], it may make sense to extend the `_kdf` function to
        support that case, or `_kdf_for_keystreams` for a different structure, as
        they are the more high-level tools."""
        hkdf = HKDF(
                algorithm=self.hashfun,
                length=l,
                salt=salt,
                info=cbor.dumps(info),
                backend=_hash_backend,
                )
        expanded = hkdf.derive(ikm)
        return expanded

    def derive_keys(self, master_salt, master_secret):
        """Populate sender_key, recipient_key and common_iv from the algorithm,
        hash function and id_context already configured beforehand, and from
        the passed salt and secret."""

        self.sender_key = self._kdf(master_salt, master_secret, self.sender_id, 'Key')
        self.recipient_key = self._kdf(master_salt, master_secret, self.recipient_id, 'Key')

        self.common_iv = self._kdf(master_salt, master_secret, b"", 'IV')

    # really more of the Credentials interface

    def get_oscore_context_for(self, unprotected):
        """Return a sutiable context (most easily self) for an incoming request
        if its unprotected data (COSE_KID, COSE_KID_CONTEXT) fit its
        description. If it doesn't match, it returns None.

        The default implementation just strictly checks for whether kid and any
        kid context match (not matching if a local KID context is set but none
        is given in the request); modes like Group OSCORE can spin up aspect
        objects here.
        """
        if unprotected.get(COSE_KID, None) == self.recipient_id and unprotected.get(COSE_KID_CONTEXT, None) == self.id_context:
            return self

class ReplayWindow:
    """A regular replay window of a fixed size.

    It is implemented as an index and a bitfield (represented by an integer)
    whose least significant bit represents the seqyence number of the index,
    and a 1 indicates that a number was seen. No shenanigans around implicit
    leading ones (think floating point normalization) happen.

    >>> w = ReplayWindow(32, lambda: None)
    >>> w.initialize_empty()
    >>> w.strike_out(5)
    >>> w.is_valid(3)
    True
    >>> w.is_valid(5)
    False
    >>> w.strike_out(0)
    >>> w.strike_out(1)
    >>> w.strike_out(2)
    >>> w.is_valid(1)
    False

    Jumping ahead by the window size invalidates older numbers:

    >>> w.is_valid(4)
    True
    >>> w.strike_out(35)
    >>> w.is_valid(4)
    True
    >>> w.strike_out(36)
    >>> w.is_valid(4)
    False

    Usage safety
    ------------

    For every key, the replay window can only be initielized empty once. On
    later uses, it needs to be persisted by storing the output of
    self.persist() somewhere and loaded from that persisted data.

    It is acceptable to store persistance data in the strike_out_callback, but
    that must then ensure that the data is written (flushed to a file or
    committed to a database), but that is usually inefficient.

    Stability
    ---------

    This class is not considered for stabilization yet and an implementation
    detail of the SecurityContext implementation(s).
    """

    _index = None
    """Sequence number represented by the least significant bit of _bitfield"""
    _bitfield = None
    """Integer interpreted as a bitfield, self._size wide. A digit 1 at any bit
    indicates that the bit's index (its power of 2) plus self._index was
    already seen."""

    def __init__(self, size, strike_out_callback):
        self._size = size
        self.strike_out_callback = strike_out_callback

    def is_initialized(self):
        return self._index is not None

    def initialize_empty(self):
        self._index = 0
        self._bitfield = 0

    def initialize_from_persisted(self, persisted):
        self._index = persisted['index']
        self._bitfield = persisted['bitfield']

    def initialize_from_freshlyseen(self, seen):
        """Initialize the replay window with a particular value that is just
        being observed in a fresh (ie. generated by the peer later than any
        messages processed before state was lost here) message. This marks the
        seen sequence number and all preceding it as invalid, and and all later
        ones as valid."""
        self._index = seen
        self._bitfield = 1

    def is_valid(self, number):
        if number < self._index:
            return False
        if number >= self._index + self._size:
            return True
        return (self._bitfield >> (number - self._index)) & 1 == 0

    def strike_out(self, number):
        if not self.is_valid(number):
            raise ValueError("Sequence number is not valid any more and "
                    "thus can't be removed from the window")
        overshoot = number - (self._index + self._size - 1)
        if overshoot > 0:
            self._index += overshoot
            self._bitfield >>= overshoot
        assert self.is_valid(number), "Sequence number was not valid before strike-out"
        self._bitfield |= 1 << (number - self._index)

        self.strike_out_callback()

    def persist(self):
        """Return a dict containing internal state which can be passed to init
        to recreated the replay window."""

        return {'index': self._index, 'bitfield': self._bitfield}

class FilesystemSecurityContext(CanProtect, CanUnprotect, SecurityContextUtils):
    """Security context stored in a directory as distinct files containing
    containing

    * Master secret, master salt, sender and recipient ID,
      optionally algorithm, the KDF hash function, and replay window size
      (settings.json and secrets.json, where the latter is typically readable
      only for the user)
    * sequence numbers and replay windows (sequence.json, the only file the
      process needs write access to)

    The static parameters can all either be placed in settings.json or
    secrets.json, but must not be present in both; the presence of either file
    is sufficient.

    .. warning::

        Security contexts must never be copied around and used after another
        copy was used. They should only ever be moved, and if they are copied
        (eg. as a part of a system backup), restored contexts must not be used
        again; they need to be replaced with freshly created ones.

    An additional file named `lock` is created to prevent the accidental use of
    a context by to concurrent programs.

    Note that the sequence number file is updated in an atomic fashion which
    requires file creation privileges in the directory. If privilege separation
    between settings/key changes and sequence number changes is desired, one
    way to achieve that on Linux is giving the aiocoap process's user group
    write permissions on the directory and setting the sticky bit on the
    directory, thus forbidding the user to remove the settings/secret files not
    owned by him.

    Writes due to sent sequence numbers are reduced by applying a variation on
    the mechanism of RFC8613 Appendix B.1.1 (incrementing the persisted sender
    seqence number in steps of `k`). That value is automatically grown from
    sequence_number_chunksize_start up to sequence_number_chunksize_limit.
    At runtime, the receive window is not stored but kept indeterminate. In
    case of an abnormal shutdown, the server uses the mechanism described in
    Appendix B.1.2 to recover.
    """

    class LoadError(ValueError):
        """Exception raised with a descriptive message when trying to load a
        faulty security context"""

    def __init__(
            self,
            basedir,
            sequence_number_chunksize_start=10,
            sequence_number_chunksize_limit=10000,
            ):
        self.basedir = basedir

        self.lockfile = filelock.FileLock(os.path.join(basedir, 'lock'))
        # 0.001: Just fail if it can't be acquired
        # See https://github.com/benediktschmitt/py-filelock/issues/57
        try:
            self.lockfile.acquire(timeout=0.001)
        # see https://github.com/PyCQA/pycodestyle/issues/703
        except: # noqa: E722
            # No lock, no loading, no need to fail in __del__
            self.lockfile = None
            raise

        # Always enabled as committing to a file for every received request
        # would be a terrible burden.
        self.echo_recovery = secrets.token_bytes(8)

        try:
            self._load()
        except KeyError as k:
            raise self.LoadError("Configuration key missing: %s" % (k.args[0],))

        self.sequence_number_chunksize_start = sequence_number_chunksize_start
        self.sequence_number_chunksize_limit = sequence_number_chunksize_limit
        self.sequence_number_chunksize = sequence_number_chunksize_start

        self.sequence_number_persisted = self.sender_sequence_number

    def _load(self):
        # doesn't check for KeyError on every occasion, relies on __init__ to
        # catch that

        data = {}
        for readfile in ("secret.json", "settings.json"):
            try:
                with open(os.path.join(self.basedir, readfile)) as f:
                    filedata = json.load(f)
            except FileNotFoundError:
                continue

            for (key, value) in filedata.items():
                if key.endswith('_hex'):
                    key = key[:-4]
                    value = binascii.unhexlify(value)
                elif key.endswith('_ascii'):
                    key = key[:-6]
                    value = value.encode('ascii')

                if key in data:
                    raise self.LoadError("Datum %r present in multiple input files at %r." % (key, self.basedir))

                data[key] = value

        self.alg_aead = algorithms[data.get('algorithm', DEFAULT_ALGORITHM)]
        self.hashfun = hashfunctions[data.get('kdf-hashfun', DEFAULT_HASHFUNCTION)]

        windowsize = data.get('window', DEFAULT_WINDOWSIZE)
        if not isinstance(windowsize, int):
            raise self.LoadError("Non-integer replay window")

        self.sender_id = data['sender-id']
        self.recipient_id = data['recipient-id']

        if max(len(self.sender_id), len(self.recipient_id)) > self.alg_aead.iv_bytes - 6:
            raise self.LoadError("Sender or Recipient ID too long (maximum length %s for this algorithm)" % (self.alg_aead.iv_bytes - 6))

        master_secret = data['secret']
        master_salt = data.get('salt', b'')
        self.id_context = data.get('id-context', None)

        self.derive_keys(master_salt, master_secret)

        self.recipient_replay_window = ReplayWindow(windowsize, self._replay_window_changed)
        try:
            with open(os.path.join(self.basedir, 'sequence.json')) as f:
                sequence = json.load(f)
        except FileNotFoundError:
            self.sender_sequence_number = 0
            self.recipient_replay_window.initialize_empty()
            self.replay_window_persisted = True
        else:
            self.sender_sequence_number = int(sequence['next-to-send'])
            received = sequence['received']
            if received == "unknown":
                # The replay window will stay uninitialized, which triggers
                # Echo recovery
                self.replay_window_persisted = False
            else:
                try:
                    self.recipient_replay_window.initialize_from_persisted(received)
                except (ValueError, TypeError, KeyError):
                    # Not being particularly careful about what could go wrong: If
                    # someone tampers with the replay data, we're already in *big*
                    # trouble, of which I fail to see how it would become worse
                    # than a crash inside the application around "failure to
                    # right-shift a string" or that like; at worst it'd result in
                    # nonce reuse which tampering with the replay window file
                    # already does.
                    raise self.LoadError("Persisted replay window state was not understood")
                self.replay_window_persisted = True

    # This is called internally whenever a new sequence number is taken or
    # crossed out from the window, and blocks a lot; B.1 mode mitigates that.
    #
    # Making it async and block in a threadpool would mitigate the blocking of
    # other messages, but the more visible effect of this will be that no
    # matter if sync or async, a reply will need to wait for a file sync
    # operation to conclude.
    def _store(self):
        tmphand, tmpnam = tempfile.mkstemp(dir=self.basedir,
                prefix='.sequence-', suffix='.json', text=True)

        data = {"next-to-send": self.sequence_number_persisted}
        if not self.replay_window_persisted:
            data['received'] = 'unknown'
        else:
            data['received'] = self.recipient_replay_window.persist()

        # Using io.open (instead os.fdopen) and binary / write with encode
        # rather than dumps as that works even while the interpreter is
        # shutting down.
        #
        # This can be relaxed when there is a defined shutdown sequence for
        # security contexts that's triggered from the general context shutdown
        # -- but right now, there isn't.
        with io.open(tmphand, 'wb') as tmpfile:
            tmpfile.write(json.dumps(data).encode('utf8'))
            tmpfile.flush()
            os.fsync(tmpfile.fileno())

        os.replace(tmpnam, os.path.join(self.basedir, 'sequence.json'))

    def _replay_window_changed(self):
        if self.replay_window_persisted:
            # Just remove the sequence numbers once from the file
            self.replay_window_persisted = False
            self._store()

    def post_seqnoincrease(self):
        if self.sender_sequence_number > self.sequence_number_persisted:
            self.sequence_number_persisted += self.sequence_number_chunksize

            self.sequence_number_chunksize = min(self.sequence_number_chunksize * 2, self.sequence_number_chunksize_limit)
            # FIXME: this blocks -- see https://github.com/chrysn/aiocoap/issues/178
            self._store()

            # The = case would only happen if someone deliberately sets all
            # numbers to 1 to force persisting on every step
            assert self.sender_sequence_number <= self.sequence_number_persisted, "Using a sequence number that has been persisted already"

    def _destroy(self):
        """Release the lock file, and ensure tha he object has become
        unusable.

        If there is unpersisted state from B.1 operation, the actually used
        number and replay window gets written back to the file to allow
        resumption without wasting digits or round-trips.
        """
        # FIXME: Arrange for a more controlled shutdown through the credentials

        self.replay_window_persisted = True
        self.sequence_number_persisted = self.sender_sequence_number
        self._store()

        del self.sender_key
        del self.recipient_key

        os.unlink(self.lockfile.lock_file)
        self.lockfile.release()

        self.lockfile = None

    def __del__(self):
        if self.lockfile is not None:
            self._destroy()

class GroupContext(BaseSecurityContext):
    is_signing = True
    external_aad_is_group = True
    responses_send_kid = True

    # This is None iff the group does not support group mode
    alg_signature_enc: Optional[AeadAlgorithm]
    # This is None iff the group does not support group mode
    alg_signature: Optional[AlgorithmCountersign]
    # This is None iff the group does not support pairwise
    #
    # This is also of type AlgorithmCountersign because the staticstatic
    # function is sitting on the same type.
    alg_pairwise_key_agreement: Optional[AlgorithmCountersign]

    @abc.abstractproperty
    def private_key(self):
        """Private key used to sign outgoing messages.

        Contexts not designed to send messages may raise a RuntimeError here;
        that necessity may later go away if some more accurate class modelling
        is found."""

    @abc.abstractproperty
    def recipient_public_key(self):
        """Public key used to verify incoming messages.

        Contexts not designed to receive messages (because they'd have aspects
        for that) may raise a RuntimeError here; that necessity may later go
        away if some more accurate class modelling is found."""

class SimpleGroupContext(GroupContext, CanProtect, CanUnprotect, SecurityContextUtils):
    """A context for an OSCORE group

    This is a non-persistable version of a group context that does not support
    any group manager or rekeying; it is set up statically at startup.

    It is intended for experimentation and demos, but aims to be correct enough
    to be usable securely.
    """

    # set during initialization
    private_key = None
    sender_auth_cred = None

    def __init__(self, alg_aead, hashfun, alg_signature, alg_signature_enc, alg_pairwise_key_agreement, group_id, master_secret, master_salt, sender_id, private_key, sender_auth_cred, peers, group_manager_cred=None, cred_fmt=COSE_KCCS):
        self.sender_id = sender_id
        self.id_context = group_id
        self.private_key = private_key
        self.alg_aead = alg_aead
        self.hashfun = hashfun
        self.alg_signature = alg_signature
        self.alg_signature_enc = alg_signature_enc
        self.alg_pairwise_key_agreement = alg_pairwise_key_agreement
        self.sender_auth_cred = sender_auth_cred
        self.group_manager_cred = group_manager_cred
        self.cred_fmt = cred_fmt

        self.peers = peers.keys()
        self.recipient_public_keys = {k: self._parse_credential(v) for (k, v) in peers.items()}
        self.recipient_auth_creds = peers
        self.recipient_replay_windows = {}
        for k in self.peers:
            # no need to persist, the whole group is ephemeral
            w = ReplayWindow(32, lambda: None)
            w.initialize_empty()
            self.recipient_replay_windows[k] = w

        self.derive_keys(master_salt, master_secret)
        self.sender_sequence_number = 0

        sender_public_key = self._parse_credential(sender_auth_cred)
        if self.alg_signature.public_from_private(self.private_key) != sender_public_key:
            raise ValueError("The key in the provided sender credential does not match the private key")

    def _parse_credential(self, credential: bytes):
        """Extract the public key (in the public_key format the respective
        AlgorithmCountersign needs) from credentials. This raises a ValueError
        if the credentials do not match the group's cred_fmt, or if the
        parameters do not match those configured in the group.

        This currently discards any information that is present in the
        credential that exceeds the key. (In a future version, this could
        return both the key and extracted other data, where that other data
        would be stored with the peer this is parsed from).
        """

        if self.cred_fmt != COSE_KCCS:
            raise ValueError("Credential parsing is currently only implemented for CCSs")

        try:
            parsed = cbor.loads(credential)
        except cbor.CBORDecodeError as e:
            raise ValueError("CCS not in CBOR format") from e

        CWT_CLAIM_CNF = 8
        CWT_CNF_COSE_KEY = 1
        if (
                not isinstance(parsed, dict)
                or CWT_CLAIM_CNF not in parsed
                or not isinstance(parsed[CWT_CLAIM_CNF], dict)
                or CWT_CNF_COSE_KEY not in parsed[CWT_CLAIM_CNF]
                ):
            raise ValueError("CCS must contain a COSE Key in a CNF")

        COSE_KEY_COMMON_KTY = 1
        COSE_KTY_OKP = 1
        COSE_KTY_EC2 = 2
        COSE_KEY_COMMON_ALG = 3
        COSE_KEY_OKP_CRV = -1
        COSE_KEY_OKP_X = -2
        COSE_KEY_EC2_X = -2
        COSE_KEY_EC2_Y = -3
        # eg. {1: 1, 3: -8, -1: 6, -2: h'77 / ... / 88'}
        cose_key = parsed[CWT_CLAIM_CNF][CWT_CNF_COSE_KEY]
        if not isinstance(cose_key, dict):
            raise ValueError("COSE Key in CCS must be a map")

        if (
                cose_key.get(COSE_KEY_COMMON_KTY) == COSE_KTY_OKP
                and cose_key.get(COSE_KEY_COMMON_ALG) == Ed25519.value
                and cose_key.get(COSE_KEY_OKP_CRV) == Ed25519.curve_number
                and COSE_KEY_OKP_X in cose_key
                ):
            return cose_key[COSE_KEY_OKP_X]
        elif (
                cose_key.get(COSE_KEY_COMMON_KTY) == COSE_KTY_EC2
                and cose_key.get(COSE_KEY_COMMON_ALG) == ECDSA_SHA256_P256.value
                and COSE_KEY_EC2_X in cose_key
                and COSE_KEY_EC2_Y in cose_key
                ):
            return ECDSA_SHA256_P256().from_public_parts(
                    x=cose_key[COSE_KEY_EC2_X],
                    y=cose_key[COSE_KEY_EC2_Y],
                    )
        else:
            raise ValueError("Key type not recognized from CCS key %r" % cose_key)

        return cbor.loads(credential)[8][1][-2]

    def __repr__(self):
        return "<%s with group %r sender_id %r and %d peers>" % (
                type(self).__name__,
                self.id_context.hex(),
                self.sender_id.hex(),
                len(self.peers),
                )

    @property
    def recipient_public_key(self):
        raise RuntimeError("Group context without key indication was used for verification")

    def derive_keys(self, master_salt, master_secret):
        # FIXME unify with parent?

        self.sender_key = self._kdf(master_salt, master_secret, self.sender_id, 'Key')
        self.recipient_keys = {recipient_id: self._kdf(master_salt, master_secret, recipient_id, 'Key') for recipient_id in self.peers}

        self.common_iv = self._kdf(master_salt, master_secret, b"", 'IV')

        # but this one is new

        self.group_encryption_key = self._kdf(master_salt, master_secret, b"", "Group Encryption Key")

    def post_seqnoincrease(self):
        """No-op because it's ephemeral"""

    def context_from_response(self, unprotected_bag) -> CanUnprotect:
        # sender ID *needs to be* here -- if this were a pairwise request, it
        # would not run through here
        try:
            sender_kid = unprotected_bag[COSE_KID]
        except KeyError:
            raise DecodeError("Group server failed to send own sender KID")

        if COSE_COUNTERSIGNATURE0 in unprotected_bag:
            return _GroupContextAspect(self, sender_kid)
        else:
            return _PairwiseContextAspect(self, sender_kid)

    def get_oscore_context_for(self, unprotected):
        if unprotected.get(COSE_KID_CONTEXT, None) != self.id_context:
            return None

        kid = unprotected.get(COSE_KID, None)
        if kid in self.peers:
            if COSE_COUNTERSIGNATURE0 in unprotected:
                return _GroupContextAspect(self, kid)
            elif self.recipient_public_keys[kid] is DETERMINISTIC_KEY:
                return _DeterministicUnprotectProtoAspect(self, kid)
            else:
                return _PairwiseContextAspect(self, kid)

    # yet to stabilize...

    def pairwise_for(self, recipient_id):
        return _PairwiseContextAspect(self, recipient_id)

    def for_sending_deterministic_requests(self, deterministic_id, target_server: Optional[bytes]):
        return _DeterministicProtectProtoAspect(self, deterministic_id, target_server)

class _GroupContextAspect(GroupContext, CanUnprotect, SecurityContextUtils):
    """The concrete context this host has with a particular peer

    As all actual data is stored in the underlying groupcontext, this acts as
    an accessor to that object (which picks the right recipient key).

    This accessor is for receiving messages in group mode from a particular
    peer; it does not send (and turns into a pairwise context through
    context_for_response before it comes to that).
    """

    def __init__(self, groupcontext, recipient_id):
        self.groupcontext = groupcontext
        self.recipient_id = recipient_id

    def __repr__(self):
        return "<%s inside %r with the peer %r>" % (
                type(self).__name__,
                self.groupcontext,
                self.recipient_id.hex(),
                )

    id_context = property(lambda self: self.groupcontext.id_context)
    alg_aead = property(lambda self: self.groupcontext.alg_aead)
    alg_signature = property(lambda self: self.groupcontext.alg_signature)
    alg_signature_enc = property(lambda self: self.groupcontext.alg_signature_enc)
    alg_pairwise_key_agreement = property(lambda self: self.groupcontext.alg_pairwise_key_agreement)
    group_manager_cred = property(lambda self: self.groupcontext.group_manager_cred)
    common_iv = property(lambda self: self.groupcontext.common_iv)

    hashfun = property(lambda self: self.groupcontext.hashfun)
    group_encryption_key = property(lambda self: self.groupcontext.group_encryption_key)

    recipient_key = property(lambda self: self.groupcontext.recipient_keys[self.recipient_id])
    recipient_public_key = property(lambda self: self.groupcontext.recipient_public_keys[self.recipient_id])
    recipient_auth_cred = property(lambda self: self.groupcontext.recipient_auth_creds[self.recipient_id])
    recipient_replay_window = property(lambda self: self.groupcontext.recipient_replay_windows[self.recipient_id])

    def context_for_response(self):
        return self.groupcontext.pairwise_for(self.recipient_id)

    @property
    def sender_auth_cred(self):
        raise RuntimeError("Could relay the sender auth credential from the group context, but it shouldn't matter here")

class _PairwiseContextAspect(GroupContext, CanProtect, CanUnprotect, SecurityContextUtils):
    is_signing = False

    def __init__(self, groupcontext, recipient_id):
        self.groupcontext = groupcontext
        self.recipient_id = recipient_id

        shared_secret = self.alg_pairwise_key_agreement.staticstatic(
                self.groupcontext.private_key,
                self.groupcontext.recipient_public_keys[recipient_id]
                )

        self.sender_key = self._kdf(
                self.groupcontext.sender_key,
                (
                    self.groupcontext.sender_auth_cred
                    + self.groupcontext.recipient_auth_creds[recipient_id]
                    + shared_secret
                ),
                self.groupcontext.sender_id,
                'Key',
                )
        self.recipient_key = self._kdf(
                self.groupcontext.recipient_keys[recipient_id],
                (
                    self.groupcontext.recipient_auth_creds[recipient_id]
                    + self.groupcontext.sender_auth_cred
                    + shared_secret
                ),
                self.recipient_id,
                'Key',
                )

    def __repr__(self):
        return "<%s based on %r with the peer %r>" % (
                type(self).__name__,
                self.groupcontext,
                self.recipient_id.hex(),
                )

    # FIXME: actually, only to be sent in requests
    id_context = property(lambda self: self.groupcontext.id_context)
    alg_aead = property(lambda self: self.groupcontext.alg_aead)
    hashfun = property(lambda self: self.groupcontext.hashfun)
    alg_signature = property(lambda self: self.groupcontext.alg_signature)
    alg_signature_enc = property(lambda self: self.groupcontext.alg_signature_enc)
    alg_pairwise_key_agreement = property(lambda self: self.groupcontext.alg_pairwise_key_agreement)
    group_manager_cred = property(lambda self: self.groupcontext.group_manager_cred)
    common_iv = property(lambda self: self.groupcontext.common_iv)
    sender_id = property(lambda self: self.groupcontext.sender_id)

    recipient_auth_cred = property(lambda self: self.groupcontext.recipient_auth_creds[self.recipient_id])
    sender_auth_cred = property(lambda self: self.groupcontext.sender_auth_cred)

    recipient_replay_window = property(lambda self: self.groupcontext.recipient_replay_windows[self.recipient_id])

    # Set at initialization
    recipient_key = None
    sender_key = None

    @property
    def sender_sequence_number(self):
        return self.groupcontext.sender_sequence_number
    @sender_sequence_number.setter
    def sender_sequence_number(self, new):
        self.groupcontext.sender_sequence_number = new

    def post_seqnoincrease(self):
        self.groupcontext.post_seqnoincrease()

    # same here -- not needed because not signing
    private_key = property(post_seqnoincrease)
    recipient_public_key = property(post_seqnoincrease)

    def context_from_response(self, unprotected_bag) -> CanUnprotect:
        if unprotected_bag.get(COSE_KID, self.recipient_id) != self.recipient_id:
            raise DecodeError("Response coming from a different server than requested, not attempting to decrypt")

        if COSE_COUNTERSIGNATURE0 in unprotected_bag:
            # It'd be an odd thing to do, but it's source verified, so the
            # server hopefully has reasons to make this readable to other group
            # members.
            return _GroupContextAspect(self.groupcontext, self.recipient_id)
        else:
            return self

class _DeterministicProtectProtoAspect(CanProtect, SecurityContextUtils):
    """This implements the sending side of Deterministic Requests.

    While simialr to a _PairwiseContextAspect, it only derives the key at
    protection time, as the plain text is hashed into the key."""

    deterministic_hashfun = hashes.SHA256()

    def __init__(self, groupcontext, sender_id, target_server: Optional[bytes]):
        self.groupcontext = groupcontext
        self.sender_id = sender_id
        self.target_server = target_server

    def __repr__(self):
        return "<%s based on %r with the sender ID %r%s>" % (
                type(self).__name__,
                self.groupcontext,
                self.sender_id.hex(),
                "limited to responses from %s" % self.target_server if self.target_server is not None else ""
                )

    def new_sequence_number(self):
        return 0

    def post_seqnoincrease(self):
        pass

    def context_from_response(self, unprotected_bag):
        if self.target_server is None:
            if COSE_KID not in unprotected_bag:
                raise DecodeError("Server did not send a KID and no particular one was addressed")
        else:
            if unprotected_bag.get(COSE_KID, self.target_server) != self.target_server:
                raise DecodeError("Response coming from a different server than requested, not attempting to decrypt")

        if COSE_COUNTERSIGNATURE0 not in unprotected_bag:
            # Could just as well pass and later barf when the group context doesn't find a signature
            raise DecodeError("Response to deterministic request came from unsecure pairwise context")

        return _GroupContextAspect(self.groupcontext, unprotected_bag.get(COSE_KID, self.target_server))

    def _get_sender_key(self, outer_message, aad, plaintext, request_id):
        if outer_message.code.is_response():
            raise RuntimeError("Deterministic contexts shouldn't protect responses")

        basekey = self.groupcontext.recipient_keys[self.sender_id]

        h = hashes.Hash(self.deterministic_hashfun)
        h.update(basekey)
        h.update(aad)
        h.update(plaintext)
        request_hash = h.finalize()

        outer_message.opt.request_hash = request_hash
        outer_message.code = FETCH

        # By this time, the AADs have all been calculated already; setting this
        # for the benefit of the response parsing later
        request_id.request_hash = request_hash
        # FIXME I don't think this ever comes to bear but want to be sure
        # before removing this line (this should only be client-side)
        request_id.can_reuse_nonce = False
        # FIXME: we're still sending a h'00' PIV. Not wrong, just a wasted byte.

        return self._kdf(basekey, request_hash, self.sender_id, 'Key')

    external_aad_is_group = True

    # details needed for various operations, especially eAAD generation
    alg_aead = property(lambda self: self.groupcontext.alg_aead)
    hashfun = property(lambda self: self.groupcontext.hashfun)
    common_iv = property(lambda self: self.groupcontext.common_iv)
    id_context = property(lambda self: self.groupcontext.id_context)
    alg_signature = property(lambda self: self.groupcontext.alg_signature)

class _DeterministicUnprotectProtoAspect(CanUnprotect, SecurityContextUtils):
    """This implements the sending side of Deterministic Requests.

    While simialr to a _PairwiseContextAspect, it only derives the key at
    unprotection time, based on information given as Request-Hash."""

    # Unless None, this is the value by which the running process recognizes
    # that the second phase of a B.1.2 replay window recovery Echo option comes
    # from the current process, and thus its sequence number is fresh
    echo_recovery = None

    deterministic_hashfun = hashes.SHA256()

    class ZeroIsAlwaysValid:
        """Special-purpose replay window that accepts 0 indefinitely"""

        def is_initialized(self):
            return True

        def is_valid(self, number):
            # No particular reason to be lax here
            return number == 0

        def strike_out(self, number):
            # FIXME: I'd rather indicate here that it's a potential replay, have the
            # request_id.can_reuse_nonce = False
            # set here rather than in _post_decrypt_checks, and thus also get
            # the check for whether it's a safe method
            pass

        def persist(self):
            pass

    def __init__(self, groupcontext, recipient_id):
        self.groupcontext = groupcontext
        self.recipient_id = recipient_id

        self.recipient_replay_window = self.ZeroIsAlwaysValid()

    def __repr__(self):
        return "<%s based on %r with the recipient ID %r>" % (
                type(self).__name__,
                self.groupcontext,
                self.recipient_id.hex(),
                )

    def context_for_response(self):
        return self.groupcontext

    def _get_recipient_key(self, protected_message):
        return self._kdf(self.groupcontext.recipient_keys[self.recipient_id], protected_message.opt.request_hash, self.recipient_id, 'Key')

    def _post_decrypt_checks(self, aad, plaintext, protected_message, request_id):
        if plaintext[0] not in (GET, FETCH): # FIXME: "is safe"
            # FIXME: accept but return inner Unauthorized. (Raising Unauthorized
            # here would just create an unprotected Unauthorized, which is not
            # what's spec'd for here)
            raise ProtectionInvalid("Request was not safe")

        basekey = self.groupcontext.recipient_keys[self.recipient_id]

        h = hashes.Hash(self.deterministic_hashfun)
        h.update(basekey)
        h.update(aad)
        h.update(plaintext)
        request_hash = h.finalize()

        if request_hash != protected_message.opt.request_hash:
            raise ProtectionInvalid("Client's hash of the plaintext diverges from the actual request hash")

        # This is intended for the protection of the response, and the
        # later use in signature in the unprotect function is not happening
        # here anyway, neither is the later use for Echo requests
        request_id.request_hash = request_hash
        request_id.can_reuse_nonce = False

    external_aad_is_group = True

    # details needed for various operations, especially eAAD generation
    alg_aead = property(lambda self: self.groupcontext.alg_aead)
    hashfun = property(lambda self: self.groupcontext.hashfun)
    common_iv = property(lambda self: self.groupcontext.common_iv)
    id_context = property(lambda self: self.groupcontext.id_context)
    alg_signature = property(lambda self: self.groupcontext.alg_signature)

def verify_start(message):
    """Extract the unprotected COSE options from a
    message for the verifier to then pick a security context to actually verify
    the message. (Future versions may also report fields from both unprotected
    and protected, if the protected bag is ever used with OSCORE.).

    Call this only requests; for responses, you'll have to know the security
    context anyway, and there is usually no information to be gained."""

    _, _, unprotected, _ = CanUnprotect._extract_encrypted0(message)

    return unprotected



_getattr__ = deprecation_getattr({
        'COSE_COUNTERSINGATURE0': 'COSE_COUNTERSIGNATURE0',
        'Algorithm': 'AeadAlgorithm',
        }, globals())



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Helpers for creating server-style applications in aiocoap

Note that these are not particular to aiocoap, but are used at different places
in aiocoap and thus shared here."""

import argparse
import sys
import logging
import asyncio
import signal

from ..util.asyncio import py38args

class ActionNoYes(argparse.Action):
    """Simple action that automatically manages --{,no-}something style options"""
    # adapted from Omnifarious's code on
    # https://stackoverflow.com/questions/9234258/in-python-argparse-is-it-possible-to-have-paired-no-something-something-arg#9236426
    def __init__(self, option_strings, dest, default=True, required=False, help=None):
        assert len(option_strings) == 1, "ActionNoYes takes only one option name"
        assert option_strings[0].startswith('--'), "ActionNoYes options must start with --"
        super().__init__(['--' + option_strings[0][2:], '--no-' + option_strings[0][2:]], dest, nargs=0, const=None, default=default, required=required, help=help)
    def __call__(self, parser, namespace, values, option_string=None):
        if option_string.startswith('--no-'):
            setattr(namespace, self.dest, False)
        else:
            setattr(namespace, self.dest, True)

class AsyncCLIDaemon:
    """Helper for creating daemon-style CLI prorgrams.

    Note that this currently doesn't create a Daemon in the sense of doing a
    daemon-fork; that could be added on demand, though.

    Subclass this and implement the :meth:`start` method as an async
    function; it will be passed all the constructor's arguments.

    When all setup is complete and the program is operational, return from the
    start method.

    Implement the :meth:`shutdown` coroutine and to do cleanup; what actually
    runs your program will, if possible, call that and await its return.

    Two usage patterns for this are supported:

    * Outside of an async context, run run ``MyClass.sync_main()``, typically
      in the program's ``if __name__ == "__main__":`` section.

      In this mode, the loop that is started is configured to safely shut down
      the loop when SIGINT is received.

    * To run a subclass of this in an existing loop, start it with
      ``MyClass(...)`` (possibly passing in the loop to run it on if not already
      in an async context), and then awaiting its ``.initializing`` future. To
      stop it, await its ``.shutdown()`` method.

      Note that with this usage pattern, the :meth:`.stop()` method has no
      effect; servers that ``.stop()`` themselves need to signal their desire
      to be shut down through other channels (but that is an atypical case).
    """

    def __init__(self, *args, **kwargs):
        loop = kwargs.pop('loop', None)
        if loop is None:
            loop = asyncio.get_running_loop()
        self.__exitcode = loop.create_future()
        self.initializing = loop.create_task(
                self.start(*args, **kwargs),
                **py38args(name="Initialization of %r" % (self,))
                )

    def stop(self, exitcode):
        """Stop the operation (and exit sync_main) at the next convenience."""
        self.__exitcode.set_result(exitcode)

    @classmethod
    async def _async_main(cls, *args, **kwargs):
        """Run the application in an AsyncIO main loop, shutting down cleanly
        on keyboard interrupt.

        This is not exposed publicly as it messes with the loop, and we only do
        that with loops created in sync_main.
        """
        main = cls(*args, **kwargs)

        try:
            asyncio.get_running_loop().add_signal_handler(
                    signal.SIGTERM,
                    lambda: main.__exitcode.set_result(143),
                    )
        except NotImplementedError:
            # Impossible on win32 -- just won't make that clean of a shutdown.
            pass

        try:
            await main.initializing
            # This is the time when we'd signal setup completion by the parent
            # exiting in case of a daemon setup, or to any other process
            # management.
            logging.info("Application ready.")
            # Common options are 143 or 0
            # (<https://github.com/go-task/task/issues/75#issuecomment-339466142> and
            # <https://unix.stackexchange.com/questions/10231/when-does-the-system-send-a-sigterm-to-a-process>)
            exitcode = await main.__exitcode
        except KeyboardInterrupt:
            logging.info("Keyboard interupt received, shutting down")
            sys.exit(3)
        else:
            sys.exit(exitcode)
        finally:
            if main.initializing.done() and main.initializing.exception():
                # The exception if initializing is what we are just watching
                # fly by. No need to trigger it again, and running shutdown
                # would be even weirder.
                pass
            else:
                # May be done, then it's a no-op, or we might have received a
                # signal during startup in which case we better fetch the
                # result and shut down cleanly again
                await main.initializing

                # And no matter whether that happened during initialization
                # (which now has finished) or due to a regular signal...
                await main.shutdown()

    @classmethod
    def sync_main(cls, *args, **kwargs):
        """Run the application in an AsyncIO main loop, shutting down cleanly
        on keyboard interrupt."""
        asyncio.run(cls._async_main(*args, **kwargs))



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This module implements a MessageInterface for UDP based on a variation of
the asyncio DatagramProtocol.

This implementation strives to be correct and complete behavior while still
only using a single socket; that is, to be usable for all kinds of multicast
traffic, to support server and client behavior at the same time, and to work
correctly even when multiple IPv6 and IPv4 (using V4MAPPED style addresses)
interfaces are present, and any of the interfaces has multiple addresses.

This requires using some standardized but not necessarily widely ported
features: ``IPV6_RECVPKTINFO`` to determine
incoming packages' destination addresses (was it multicast) and to return
packages from the same address, ``IPV6_JOIN_GROUP`` for multicast
membership management and ``recvmsg`` to obtain data configured with the above
options. The need for ``AI_V4MAPPED`` and ``AI_ADDRCONFIG`` is not manifest
in the code because the latter on its own is insufficient to enable seamless
interoperability with IPv4+IPv6 servers on IPv4-only hosts; instead,
short-lived sockets are crated to assess which addresses are routable. This
should correctly deal with situations in which a client has an IPv6 ULA
assigned but no route, no matter whether the server advertises global IPv6
addresses or addresses inside that ULA. It can not deal with situations in
which the host has a default IPv6 route, but that route is not actually usable.

To the author's knowledge, there is no standardized mechanism for receiving
ICMP errors in such a setup. On Linux, ``IPV6_RECVERR`` and ``MSG_ERRQUEUE``
are used to receive ICMP errors from the socket; on other platforms, a warning
is emitted that ICMP errors are ignored. Using a :mod:`.simple6` for clients is
recommended for those when working as a client only.

Exceeding for the above error handling, no attempts are made to fall back to a
kind-of-correct or limited-functionality behavior if these options are
unavailable, for the resulting code would be hard to maintain ("``ifdef``
hell") or would cause odd bugs at users (eg. servers that stop working when an
additional IPv6 address gets assigned). If the module does not work for you,
and the options can not be added easily to your platform, consider using the
:mod:`.simple6` module instead.
"""

import asyncio
import contextvars
import socket
import ipaddress
import struct
import weakref
from collections import namedtuple

from ..message import Message
from ..numbers import constants
from .. import defaults
from .. import error
from .. import interfaces
from ..numbers import COAP_PORT
from ..util.asyncio.recvmsg import RecvmsgDatagramProtocol, create_recvmsg_datagram_endpoint
from ..util.asyncio.getaddrinfo_addrconfig import getaddrinfo_routechecked as getaddrinfo
from ..util import hostportjoin, hostportsplit
from ..util import socknumbers

"""The `struct in6_pktinfo` from RFC3542"""
_in6_pktinfo = struct.Struct("16sI")

_ipv6_unspecified = socket.inet_pton(socket.AF_INET6, '::')
_ipv4_unspecified = socket.inet_pton(socket.AF_INET6, '::ffff:0.0.0.0')

class InterfaceOnlyPktinfo(bytes):
    """A thin wrapper over bytes that represent a pktinfo built just to select
    an outgoing interface.

    This must not be treated any different than a regular pktinfo, and is just
    tagged for better debug output. (Ie. if this is replaced everywhere with
    plain `bytes`, things must still work)."""

class UDP6EndpointAddress(interfaces.EndpointAddress):
    """Remote address type for :class:`MessageInterfaceUDP6`. Remote address is
    stored in form of a socket address; local address can be roundtripped by
    opaque pktinfo data.

    For purposes of equality (and thus hashing), the local address is *not*
    checked. Neither is the scopeid that is part of the socket address.

    >>> interface = type("FakeMessageInterface", (), {})
    >>> if1_name = socket.if_indextoname(1)
    >>> local = UDP6EndpointAddress(socket.getaddrinfo('127.0.0.1', 5683, type=socket.SOCK_DGRAM, family=socket.AF_INET6, flags=socket.AI_V4MAPPED)[0][-1], interface)
    >>> local.is_multicast
    False
    >>> local.hostinfo
    '127.0.0.1'
    >>> all_coap_link1 = UDP6EndpointAddress(socket.getaddrinfo('ff02:0:0:0:0:0:0:fd%1', 1234, type=socket.SOCK_DGRAM, family=socket.AF_INET6)[0][-1], interface)
    >>> all_coap_link1.is_multicast
    True
    >>> all_coap_link1.hostinfo == '[ff02::fd%{}]:1234'.format(if1_name)
    True
    >>> all_coap_site = UDP6EndpointAddress(socket.getaddrinfo('ff05:0:0:0:0:0:0:fd', 1234, type=socket.SOCK_DGRAM, family=socket.AF_INET6)[0][-1], interface)
    >>> all_coap_site.is_multicast
    True
    >>> all_coap_site.hostinfo
    '[ff05::fd]:1234'
    >>> all_coap4 = UDP6EndpointAddress(socket.getaddrinfo('224.0.1.187', 5683, type=socket.SOCK_DGRAM, family=socket.AF_INET6, flags=socket.AI_V4MAPPED)[0][-1], interface)
    >>> all_coap4.is_multicast
    True
    """

    def __init__(self, sockaddr, interface, *, pktinfo=None):
        self.sockaddr = sockaddr
        self.pktinfo = pktinfo
        self._interface = weakref.ref(interface)

    scheme = 'coap'

    interface = property(lambda self: self._interface())

    def __hash__(self):
        return hash(self.sockaddr[:-1])

    def __eq__(self, other):
        return self.sockaddr[:-1] == other.sockaddr[:-1]

    def __repr__(self):
        return "<%s %s%s>" % (type(self).__name__, self.hostinfo, " (locally %s)" % self._repr_pktinfo() if self.pktinfo is not None else "")

    @staticmethod
    def _strip_v4mapped(address):
        """Turn anything that's a valid input to ipaddress.IPv6Address into a
        user-friendly string that's either an IPv6 or an IPv4 address.

        This also compresses (normalizes) the IPv6 address as a convenient side
        effect."""
        address = ipaddress.IPv6Address(address)
        mapped = address.ipv4_mapped
        if mapped is not None:
            return str(mapped)
        return str(address)

    def _plainaddress(self):
        """Return the IP adress part of the sockaddr in IPv4 notation if it is
        mapped, otherwise the plain v6 address including the interface
        identifier if set."""

        if self.sockaddr[3] != 0:
            try:
                scopepart = "%" + socket.if_indextoname(self.sockaddr[3])
            except Exception: # could be an OS error, could just be that there is no function of this name, as it is on Android
                scopepart = "%" + str(self.sockaddr[3])
        else:
            scopepart = ""
        if '%' in self.sockaddr[0]:
            # Fix for Python 3.6 and earlier that reported the scope information
            # in the IP literal (3.7 consistently expresses it in the tuple slot 3)
            scopepart = ""
        return self._strip_v4mapped(self.sockaddr[0]) + scopepart

    def _repr_pktinfo(self):
        """What repr(self.pktinfo) would be if that were not a plain untyped bytestring"""
        addr, interface = _in6_pktinfo.unpack_from(self.pktinfo)
        if interface == 0:
            interface = ""
        else:
            try:
                interface = "%" + socket.if_indextoname(interface)
            except Exception as e:
                interface = "%%%d(%s)" % (interface, e)

        return "%s%s" % (self._strip_v4mapped(addr), interface)

    def _plainaddress_local(self):
        """Like _plainaddress, but on the address in the pktinfo. Unlike
        _plainaddress, this does not contain the interface identifier."""

        addr, interface = _in6_pktinfo.unpack_from(self.pktinfo)

        return self._strip_v4mapped(addr)

    @property
    def netif(self):
        """Textual interface identifier of the explicitly configured remote
        interface, or the interface identifier reported in an incoming
        link-local message. None if not set."""
        index = self.sockaddr[3]
        return socket.if_indextoname(index) if index else None

    @property
    def hostinfo(self):
        port = self.sockaddr[1]
        if port == COAP_PORT:
            port = None

        # plainaddress: don't assume other applications can deal with v4mapped addresses
        return hostportjoin(self._plainaddress(), port)

    @property
    def hostinfo_local(self):
        host = self._plainaddress_local()
        port = self.interface._local_port()
        if port == 0:
            raise ValueError("Local port read before socket has bound itself")
        if port == COAP_PORT:
            port = None
        return hostportjoin(host, port)

    @property
    def uri_base(self):
        return 'coap://' + self.hostinfo

    @property
    def uri_base_local(self):
        return 'coap://' + self.hostinfo_local

    @property
    def is_multicast(self):
        return ipaddress.ip_address(self._plainaddress().split('%', 1)[0]).is_multicast

    @property
    def is_multicast_locally(self):
        return ipaddress.ip_address(self._plainaddress_local()).is_multicast

    def as_response_address(self):
        if not self.is_multicast_locally:
            return self

        # Create a copy without pktinfo, as responses to messages received to
        # multicast addresses can not have their request's destination address
        # as source address
        return type(self)(self.sockaddr, self.interface)

    @property
    def blockwise_key(self):
        return (self.sockaddr, self.pktinfo)


class SockExtendedErr(namedtuple("_SockExtendedErr", "ee_errno ee_origin ee_type ee_code ee_pad ee_info ee_data")):
    _struct = struct.Struct("IbbbbII")
    @classmethod
    def load(cls, data):
        # unpack_from: recvmsg(2) says that more data may follow
        return cls(*cls._struct.unpack_from(data))

class MessageInterfaceUDP6(RecvmsgDatagramProtocol, interfaces.MessageInterface):
    def __init__(self, ctx: interfaces.MessageManager, log, loop):
        self._ctx = ctx
        self.log = log
        self.loop = loop

        self._shutting_down = None #: Future created and used in the .shutdown() method.

        self.ready = asyncio.get_running_loop().create_future() #: Future that gets fullfilled by connection_made (ie. don't send before this is done; handled by ``create_..._context``

        # This is set while a send is underway to determine in the
        # error_received call site whom we were actually sending something to.
        # This is a workaround for the abysmal error handling unconnected
        # sockets have :-/
        #
        # FIXME: Figure out whether aiocoap can at all support a context being
        # used with multiple aiocoap contexts, and if not, raise an error early
        # rather than just-in-case doing extra stuff here.
        self._remote_being_sent_to = contextvars.ContextVar('_remote_being_sent_to', default=None)

    def _local_port(self):
        # FIXME: either raise an error if this is 0, or send a message to self
        # to force the OS to decide on a port. Right now, this reports wrong
        # results while the first message has not been sent yet.
        return self.transport.get_extra_info('socket').getsockname()[1]

    @classmethod
    async def _create_transport_endpoint(cls, sock, ctx: interfaces.MessageManager, log, loop, multicast=[]):
        try:
            sock.setsockopt(socket.IPPROTO_IPV6, socknumbers.IPV6_RECVPKTINFO, 1)
        except NameError:
            raise RuntimeError("RFC3542 PKTINFO flags are unavailable, unable to create a udp6 transport.")
        if socknumbers.HAS_RECVERR:
            sock.setsockopt(socket.IPPROTO_IPV6, socknumbers.IPV6_RECVERR, 1)
            # i'm curious why this is required; didn't IPV6_V6ONLY=0 already make
            # it clear that i don't care about the ip version as long as everything looks the same?
            sock.setsockopt(socket.IPPROTO_IP, socknumbers.IP_RECVERR, 1)
        else:
            log.warning("Transport udp6 set up on platform without RECVERR capability. ICMP errors will be ignored.")

        for (address, interface) in sum(map(
                    # Expand shortcut of "interface name means default CoAP all-nodes addresses"
                    lambda i: [(a, i) for a in constants.MCAST_ALL] if isinstance(i, str) else [i],
                    multicast
                ), []):
            address = ipaddress.ip_address(address)
            interface = socket.if_nametoindex(interface)

            if isinstance(address, ipaddress.IPv4Address):
                s = struct.pack('4s4si',
                        address.packed,
                        socket.inet_aton("0.0.0.0"), interface)
                try:
                    sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, s)
                except OSError:
                    log.warning("Could not join IPv4 multicast group")

            elif isinstance(address, ipaddress.IPv6Address):
                s = struct.pack('16si',
                        address.packed,
                        interface)
                try:
                    sock.setsockopt(socket.IPPROTO_IPV6,
                            socket.IPV6_JOIN_GROUP, s)
                except OSError:
                    log.warning("Could not join IPv6 multicast group")

            else:
                raise RuntimeError("Unknown address format")

        transport, protocol = await create_recvmsg_datagram_endpoint(loop,
                lambda: cls(ctx, log=log, loop=loop),
                sock=sock)

        await protocol.ready

        return protocol

    @classmethod
    async def create_client_transport_endpoint(cls, ctx: interfaces.MessageManager, log, loop):
        sock = socket.socket(family=socket.AF_INET6, type=socket.SOCK_DGRAM)
        sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)

        return await cls._create_transport_endpoint(sock, ctx, log, loop)

    @classmethod
    async def create_server_transport_endpoint(cls, ctx: interfaces.MessageManager, log, loop, bind, multicast):
        bind = bind or ('::', None)
        # Interpret None as 'default port', but still allow to bind to 0 for
        # servers that want a random port (eg. when the service URLs are
        # advertised out-of-band anyway, or in LwM2M clients)
        bind = (bind[0], COAP_PORT if bind[1] is None else bind[1])

        # The later bind() does most of what getaddr info usually does
        # (including resolving names), but is missing out subtly: It does not
        # populate the zone identifier of an IPv6 address, making it impossible
        # without a getaddrinfo (or manual mapping of the name to a number) to
        # bind to a specific link-local interface
        try:
            addriter = getaddrinfo(
                    loop,
                    log,
                    bind[0],
                    bind[1],
                    )
            try:
                bind = await addriter.__anext__()
            except StopAsyncIteration:
                raise RuntimeError("getaddrinfo returned zero-length list rather than erring out")
        except socket.gaierror:
            raise error.ResolutionError("No local bindable address found for %s" % bind[0])

        try:
            additional = await addriter.__anext__()
        except StopAsyncIteration:
            pass
        except Exception as e:
            log.error("Ignoring exception raised when checking for additional addresses that match the bind address", exc_info=e)
        else:
            log.warning("Multiple addresses to bind to, only selecting %r and discarding %r and any later", bind, additional)

        sock = socket.socket(family=socket.AF_INET6, type=socket.SOCK_DGRAM)
        if defaults.has_reuse_port():
            # I doubt that there is any platform that supports RECVPKTINFO but
            # not REUSEPORT, but why take chances.
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
        sock.bind(bind)

        return (await cls._create_transport_endpoint(sock, ctx, log, loop, multicast))

    async def shutdown(self):
        self._shutting_down = asyncio.get_running_loop().create_future()

        self.transport.close()

        await self._shutting_down

        del self._ctx

    def send(self, message):
        ancdata = []
        if message.remote.pktinfo is not None:
            ancdata.append((socket.IPPROTO_IPV6, socknumbers.IPV6_PKTINFO,
                message.remote.pktinfo))
        assert self._remote_being_sent_to.get(None) is None, "udp6.MessageInterfaceUDP6.send was reentered in a single task"
        self._remote_being_sent_to.set(message.remote)
        try:
            self.transport.sendmsg(message.encode(), ancdata, 0, message.remote.sockaddr)
        finally:
            self._remote_being_sent_to.set(None)

    async def recognize_remote(self, remote):
        return isinstance(remote, UDP6EndpointAddress) and \
                remote.interface == self

    async def determine_remote(self, request):
        if request.requested_scheme not in ('coap', None):
            return None

        if request.unresolved_remote is not None:
            host, port = hostportsplit(request.unresolved_remote)
            port = port or COAP_PORT
        elif request.opt.uri_host:
            host = request.opt.uri_host
            if host.startswith('[') and host.endswith(']'):
                host = host[1:-1]
            port = request.opt.uri_port or COAP_PORT
        else:
            raise ValueError("No location found to send message to (neither in .opt.uri_host nor in .remote)")

        # Take aside the zone identifier. While it can pass through getaddrinfo
        # in some situations (eg. 'fe80::1234%eth0' will give 'fe80::1234'
        # scope eth0, and similar for ff02:: addresses), in others (eg. ff05::)
        # it gives 'Name or service not known'.

        if '%' in host:
            host, zone = host.split('%', 1)
            try:
                zone = socket.if_nametoindex(zone)
            except OSError:
                raise error.ResolutionError("Invalid zone identifier %s" % zone)
        else:
            zone = None

        try:
            # Note that this is our special addrinfo that ensures there is a
            # route.
            ip, port, flowinfo, scopeid = await getaddrinfo(
                    self.loop,
                    self.log,
                    host,
                    port,
                    ).__anext__()
        except socket.gaierror:
            raise error.ResolutionError("No address information found for requests to %r" % host)

        if zone is not None:
            # Still trying to preserve the information returned (libc can't do
            # it as described at
            # <https://unix.stackexchange.com/questions/174767/ipv6-zone-id-in-etc-hosts>)
            # in case something sane does come out of that.
            if scopeid != 0 and scopeid != zone:
                self.log.warning("Resolved address of %s came with zone ID %d whereas explicit ID %d takes precedence", host, scopeid, zone)
            scopeid = zone

        # We could be done here and return UDP6EndpointAddress(the reassembled
        # sockaddr, self), but:
        #
        # Linux (unlike FreeBSD) takes the sockaddr's scope ID only for
        # link-local scopes (as per ipv6(7), and discards it otherwise. It does
        # need the information of the selected interface, though, in order to
        # pick the right outgoing interface. Thus, we provide it in the local
        # portion.

        if scopeid:
            # "Any" does not include "even be it IPv4" -- the underlying family
            # unfortunately needs to be set, or Linux will refuse to send.
            if ipaddress.IPv6Address(ip).ipv4_mapped is None:
                local_source = _ipv6_unspecified
            else:
                local_source = _ipv4_unspecified
            local = InterfaceOnlyPktinfo(_in6_pktinfo.pack(local_source, scopeid))
        else:
            local = None

        sockaddr = ip, port, flowinfo, scopeid
        return UDP6EndpointAddress(sockaddr, self, pktinfo=local)

    #
    # implementing the typical DatagramProtocol interfaces.
    #
    # note from the documentation: we may rely on connection_made to be called
    # before datagram_received -- but sending immediately after context
    # creation will still fail

    def connection_made(self, transport):
        """Implementation of the DatagramProtocol interface, called by the transport."""
        self.ready.set_result(True)
        self.transport = transport

    def datagram_msg_received(self, data, ancdata, flags, address):
        """Implementation of the RecvmsgDatagramProtocol interface, called by the transport."""
        pktinfo = None
        for cmsg_level, cmsg_type, cmsg_data in ancdata:
            if cmsg_level == socket.IPPROTO_IPV6 and cmsg_type == socknumbers.IPV6_PKTINFO:
                pktinfo = cmsg_data
            else:
                self.log.info("Received unexpected ancillary data to recvmsg: level %d, type %d, data %r", cmsg_level, cmsg_type, cmsg_data)
        if pktinfo is None:
            self.log.warning("Did not receive requested pktinfo ancdata on message from %s", address)
        try:
            message = Message.decode(data, UDP6EndpointAddress(address, self, pktinfo=pktinfo))
        except error.UnparsableMessage:
            self.log.warning("Ignoring unparsable message from %s", address)
            return

        try:
            self._ctx.dispatch_message(message)
        except BaseException as exc:
            # Catching here because util.asyncio.recvmsg inherits
            # _SelectorDatagramTransport's bad handling of callback errors;
            # this is the last time we have a log at hand.
            self.log.error("Exception raised through dispatch_message: %s", exc, exc_info=exc)
            raise

    def datagram_errqueue_received(self, data, ancdata, flags, address):
        assert flags == socknumbers.MSG_ERRQUEUE, "Received non-error data through the errqueue"
        pktinfo = None
        errno = None
        for cmsg_level, cmsg_type, cmsg_data in ancdata:
            assert cmsg_level == socket.IPPROTO_IPV6, "Received non-IPv6 protocol through the errqueue"
            if cmsg_type == socknumbers.IPV6_RECVERR:
                extended_err = SockExtendedErr.load(cmsg_data)
                self.log.debug("Socket error recevied, details: %s", extended_err)
                errno = extended_err.ee_errno
            elif cmsg_level == socket.IPPROTO_IPV6 and cmsg_type == socknumbers.IPV6_PKTINFO:
                pktinfo = cmsg_data
            else:
                self.log.info("Received unexpected ancillary data to recvmsg errqueue: level %d, type %d, data %r", cmsg_level, cmsg_type, cmsg_data)
        remote = UDP6EndpointAddress(address, self, pktinfo=pktinfo)

        # not trying to decode a message from data -- that works for
        # "connection refused", doesn't work for "no route to host", and
        # anyway, when an icmp error comes back, everything pending from that
        # port should err out.

        try:
            self._ctx.dispatch_error(OSError(errno, "received through errqueue"), remote)
        except BaseException as exc:
            # Catching here because util.asyncio.recvmsg inherits
            # _SelectorDatagramTransport's bad handling of callback errors;
            # this is the last time we have a log at hand.
            self.log.error("Exception raised through dispatch_error: %s", exc, exc_info=exc)
            raise

    def error_received(self, exc):
        """Implementation of the DatagramProtocol interface, called by the transport."""

        remote = self._remote_being_sent_to.get()

        if remote is None:
            # TODO: Are there any errors left that wind up here? (sending to 127.0.0.0 takes the later half of this function)
            self.log.error("Error received in situation with no way to to determine which sending caused the error; this should be accompanied by an error in another code path: %s", exc)
            return

        try:
            self._ctx.dispatch_error(exc, remote)
        except BaseException as exc:
            # Catching here because util.asyncio.recvmsg inherits
            # _SelectorDatagramTransport's bad handling of callback errors;
            # this is the last time we have a log at hand.
            self.log.error("Exception raised through dispatch_error: %s", exc, exc_info=exc)
            raise

    def connection_lost(self, exc):
        # TODO better error handling -- find out what can cause this at all
        # except for a shutdown
        if exc is not None:
            self.log.error("Connection lost: %s", exc)

        if self._shutting_down is None:
            self.log.error("Connection loss was not expected.")
        else:
            self._shutting_down.set_result(None)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Tools not directly related with CoAP that are needed to provide the API

These are only part of the stable API to the extent they are used by other APIs
-- for example, you can use the type constructor of :class:`ExtensibleEnumMeta`
when creating an :class:`aiocoap.numbers.optionnumbers.OptionNumber`, but don't
expect it to be usable in a stable way for own extensions.

Most functions are available in submodules; some of them may only have
components that are exclusively used internally and never part of the public
API even in the limited fashion stated above.

.. toctree::
    :glob:

    aiocoap.util.*
"""

import urllib.parse

class ExtensibleEnumMeta(type):
    """Metaclass for ExtensibleIntEnum, see there for detailed explanations"""
    def __init__(self, name, bases, dict):
        self._value2member_map_ = {}
        for k, v in dict.items():
            if k.startswith('_'):
                continue
            if callable(v):
                continue
            if isinstance(v, property):
                continue
            if isinstance(v, classmethod):
                continue
            instance = self(v)
            instance.name = k
            setattr(self, k, instance)
        type.__init__(self, name, bases, dict)

    def __call__(self, value):
        if isinstance(value, self):
            return value
        if value not in self._value2member_map_:
            self._value2member_map_[value] = super(ExtensibleEnumMeta, self).__call__(value)
        return self._value2member_map_[value]

class ExtensibleIntEnum(int, metaclass=ExtensibleEnumMeta):
    """Similar to Python's enum.IntEnum, this type can be used for named
    numbers which are not comprehensively known, like CoAP option numbers."""

    def __repr__(self):
        return '<%s %d%s>' % (type(self).__name__, self, ' "%s"' % self.name if hasattr(self, "name") else "")

    def __str__(self):
        return self.name if hasattr(self, "name") else int.__str__(self)

    def _repr_html_(self):
        import html
        if hasattr(self, "name"):
            return f'<abbr title="{html.escape(type(self).__name__)} {int(self)}">{html.escape(self.name)}</abbr>'
        else:
            return f'<abbr title="Unknown {html.escape(type(self).__name__)}">{int(self)}</abbr>'

def hostportjoin(host, port=None):
    """Join a host and optionally port into a hostinfo-style host:port
    string

    >>> hostportjoin('example.com')
    'example.com'
    >>> hostportjoin('example.com', 1234)
    'example.com:1234'
    >>> hostportjoin('127.0.0.1', 1234)
    '127.0.0.1:1234'

    This is lax with respect to whether host is an IPv6 literal in brackets or
    not, and accepts either form; IP-future literals that do not contain a
    colon must be already presented in their bracketed form:

    >>> hostportjoin('2001:db8::1')
    '[2001:db8::1]'
    >>> hostportjoin('2001:db8::1', 1234)
    '[2001:db8::1]:1234'
    >>> hostportjoin('[2001:db8::1]', 1234)
    '[2001:db8::1]:1234'
    """
    if ':' in host and not (host.startswith('[') and host.endswith(']')):
        host = '[%s]' % host

    if port is None:
        hostinfo = host
    else:
        hostinfo = "%s:%d" % (host, port)
    return hostinfo

def hostportsplit(hostport):
    """Like urllib.parse.splitport, but return port as int, and as None if not
    given. Also, it allows giving IPv6 addresses like a netloc:

    >>> hostportsplit('foo')
    ('foo', None)
    >>> hostportsplit('foo:5683')
    ('foo', 5683)
    >>> hostportsplit('[::1%eth0]:56830')
    ('::1%eth0', 56830)
    """

    pseudoparsed = urllib.parse.SplitResult(None, hostport, None, None, None)
    try:
        return pseudoparsed.hostname, pseudoparsed.port
    except ValueError:
        if '[' not in hostport and hostport.count(':') > 1:
            raise ValueError("Could not parse network location. "
                "Beware that when IPv6 literals are expressed in URIs, they "
                "need to be put in square brackets to distinguish them from "
                "port numbers.")
        raise

def quote_nonascii(s):
    """Like urllib.parse.quote, but explicitly only escaping non-ascii characters.

    This function is deprecated due to it use of the irrelevant "being an ASCII
    character" property (when instead RFC3986 productions like "unreserved"
    should be used), and due for removal when aiocoap's URI processing is
    overhauled the next time.
    """

    return "".join(chr(c) if c <= 127 else "%%%02X" % c for c in s.encode('utf8'))

class Sentinel:
    """Class for sentinel that can only be compared for identity. No efforts
    are taken to make these singletons; it is up to the users to always refer
    to the same instance, which is typically defined on module level."""
    def __init__(self, label):
        self._label = label

    def __repr__(self):
        return '<%s>' % self._label

def deprecation_getattr(_deprecated_aliases: dict, _globals: dict):
    """Factory for a module-level ``__getattr__`` function

    This creates deprecation warnings whenever a module level item by one of
    the keys of the alias dict is accessed by its old name rather than by its
    new name (which is in the values):

    >>> FOOBAR = 42
    >>>
    >>> __getattr__ = deprecation_getattr({'FOOBRA': 'FOOBAR'}, globals())
    """
    def __getattr__(name):
        if name in _deprecated_aliases:
            modern = _deprecated_aliases[name]
            from warnings import warn
            warn(f"{name} is deprecated, use {modern} instead", DeprecationWarning,
                    stacklevel=2)
            return _globals[modern]
        raise AttributeError(f"module {__name__} has no attribute {name}")
    return __getattr__



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Container module for command line utilities bundled with aiocoap.

These modules are not considered to be a part of the aioCoAP API, and are thus
subject to change even when the project reaches a stable version number. If you
want to use any of that infrastructure, please file a feature request for
stabilization in the project's issue tracker.

The tools themselves are documented in :doc:`/tools`.
"""



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

from .. import socknumbers

from asyncio import BaseProtocol
from asyncio.transports import BaseTransport

class RecvmsgDatagramProtocol(BaseProtocol):
    """Callback interface similar to asyncio.DatagramProtocol, but dealing with
    recvmsg data."""

    def datagram_msg_received(self, data, ancdata, flags, address):
        """Called when some datagram is received."""

    def datagram_errqueue_received(self, data, ancdata, flags, address):
        """Called when some data is received from the error queue"""

    def error_received(self, exc):
        """Called when a send or receive operation raises an OSError."""

def _set_result_unless_cancelled(fut, result):
    """Helper setting the result only if the future was not cancelled."""
    if fut.cancelled():
        return
    fut.set_result(result)

class RecvmsgSelectorDatagramTransport(BaseTransport):
    """A simple loop-independent transport that largely mimicks
    DatagramTransport but interfaces a RecvmsgSelectorDatagramProtocol.

    This does not implement any flow control, based on the assumption that it's
    not needed, for CoAP has its own flow control mechanisms."""

    max_size = 4096  # Buffer size passed to recvmsg() -- should suffice for a full MTU package and ample ancdata

    def __init__(self, loop, sock, protocol, waiter):
        super().__init__(extra={'socket': sock})
        self.__sock = sock
        # Persisted outside of sock because when GC breaks a reference cycle,
        # it can happen that the sock gets closed before this; we have to hope
        # that no new file gets opened and registered in the meantime.
        self.__sock_fileno = sock.fileno()
        self._loop = loop
        self._protocol = protocol

        loop.call_soon(protocol.connection_made, self)
        # only start reading when connection_made() has been called
        import weakref
        # We could add error handling in here like this:
        # ```
        # self = s()
        # if self is None or self.__sock is None:
        #     # The read event happened briefly before .close() was called,
        #     # but late enough that the caller of close did not yield to let
        #     # the event out; when remove_reader was then called, the
        #     # pending event was not removed, so it fires now that the
        #     # socket is already closed. (Depending on the GC's whims, self
        #     # may or may not have been GC'd, but if it wasn't yet, the
        #     # closed state is indicated by the lack of a __sock.
        #     #
        #     # Thus, silently (preferably with an ICMP error, but really
        #     # can't do that)...
        #     return
        # ```
        # That was done tentatively while debugging errors flying out of
        # _read_ready, but it turned out that this was not the actual error
        # source. Thus, I'm not adding the handler and assuming that close's
        # remove_reader is not racing against callbacks, and thus that s() is
        # always valid while the transport is around (and the weakref is really
        # only used to break up the reference cycles to ensure the GC is not
        # needed here).
        def rr(s=weakref.ref(self)):
            s()._read_ready()
        loop.call_soon(loop.add_reader, self.__sock_fileno, rr)
        loop.call_soon(_set_result_unless_cancelled, waiter, None)

    def close(self):
        if self.__sock is None:
            return

        if not self._loop.is_closed():
            self._loop.call_soon(self._protocol.connection_lost, None)

        self._loop.remove_reader(self.__sock_fileno)
        self.__sock.close()
        self.__sock = None
        self._protocol = None
        self._loop = None

    def __del__(self):
        if self.__sock is not None:
            self.close()

    def _read_ready(self):
        if socknumbers.HAS_RECVERR:
            try:
                data, ancdata, flags, addr = self.__sock.recvmsg(self.max_size, 1024, socknumbers.MSG_ERRQUEUE)
            except (BlockingIOError, InterruptedError):
                pass
            except OSError as exc:
                if repr(exc) == "OSError('received malformed or improperly truncated ancillary data',)":
                    pass # workaround for https://bitbucket.org/pypy/pypy/issues/2649/recvmsg-with-empty-err-queue-raises-odd
                else:
                    self._protocol.error_received(exc)
            except Exception as exc:
                self._fatal_error(exc, 'Fatal read error on datagram transport')
            else:
                self._protocol.datagram_errqueue_received(data, ancdata, flags, addr)

        # copied and modified from _SelectorDatagramTransport
        try:
            data, ancdata, flags, addr = self.__sock.recvmsg(self.max_size, 1024) # TODO: find a way for the application to tell the trensport how much data is expected
        except (BlockingIOError, InterruptedError):
            pass
        except OSError as exc:
            self._protocol.error_received(exc)
        except Exception as exc:
            self._fatal_error(exc, 'Fatal read error on datagram transport')
        else:
            self._protocol.datagram_msg_received(data, ancdata, flags, addr)

    def sendmsg(self, data, ancdata, flags, address):
        try:
            self.__sock.sendmsg((data,), ancdata, flags, address)
            return
        except OSError as exc:
            self._protocol.error_received(exc)
            return
        except Exception as exc:
            self._fatal_error(exc,
                              'Fatal write error on datagram transport')
            return

async def create_recvmsg_datagram_endpoint(loop, factory, sock):
    """Create a datagram connection that uses recvmsg rather than recvfrom, and
    a RecvmsgDatagramProtocol protocol type.

    This is used like the create_datagram_endpoint method of an asyncio loop,
    but implemented in a generic way using the loop's add_reader method; thus,
    it's not a method of the loop but an independent function.

    Due to the way it is used in aiocoap, socket is not an optional argument
    here; it could be were this module ever split off into a standalone
    package.
    """
    sock.setblocking(False)

    protocol = factory()
    waiter = loop.create_future()
    transport = RecvmsgSelectorDatagramTransport(
            loop, sock, protocol, waiter)

    try:
        await waiter
    # see https://github.com/PyCQA/pycodestyle/issues/703
    except: # noqa: E722
        transport.close()
        raise

    return transport, protocol



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Helpers around content types

This uses the terminology clarified in 1_, and primarily deals with content
types in their usual string representation.

Unless content types get used a lot more in aiocoap, this provides only
accessors to some of their relevant properties, without aiming to build
semantically accessible objects to encapsulate them.

.. _1: https://tools.ietf.org/html/draft-bormann-core-media-content-type-format-01"""

def categorize(contenttype: str):
    """Return 'cbor', 'json' or 'link-format' if the content type indicates it
    is that format itself or derived from it."""

    media_type, *_ = contenttype.split(';')
    _, _, subtype = media_type.partition('/')

    if subtype == 'cbor-seq' or subtype.endswith('+cbor-seq'):
        return 'cbor-seq'

    if subtype == 'cbor' or subtype.endswith('+cbor'):
        return 'cbor'

    if subtype == 'json' or subtype.endswith('+json'):
        return 'json'

    if media_type == 'application/link-format':
        return 'link-format'

    return None
